from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

import pytest

from dissect.target.helpers.regutil import RegistryHive, VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.datetime import WindowsDateTimePlugin, c_tz, parse_systemtime_transition
from dissect.target.plugins.os.windows.locale import WindowsLocalePlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_windows_datetime(target_win_tzinfo: Target) -> None:
    target_win_tzinfo.add_plugin(WindowsDateTimePlugin)

    # Easter Island has a flipped DST to Amsterdam
    assert target_win_tzinfo.datetime.tzinfo.display == "(UTC-06:00) Easter Island"
    assert target_win_tzinfo.datetime.tzinfo.dlt_name == "Easter Island Daylight Time"
    assert target_win_tzinfo.datetime.tzinfo.std_name == "Easter Island Standard Time"
    assert 2019 in target_win_tzinfo.datetime.tzinfo.dynamic_dst

    naive_dt_may = datetime.datetime(2019, 5, 4, 12, 0, 0)  # noqa: DTZ001
    naive_dt_march = datetime.datetime(2019, 3, 4, 12, 0, 0)  # noqa: DTZ001

    local_east_st_dt = target_win_tzinfo.datetime.local(naive_dt_may)
    local_east_dt_dt = target_win_tzinfo.datetime.local(naive_dt_march)

    assert local_east_st_dt.tzinfo == target_win_tzinfo.datetime.tzinfo
    assert local_east_dt_dt.tzinfo == target_win_tzinfo.datetime.tzinfo
    assert str(local_east_st_dt) == "2019-05-04 12:00:00-06:00"
    assert str(local_east_dt_dt) == "2019-03-04 12:00:00-05:00"
    assert target_win_tzinfo.datetime.tzinfo.tzname(local_east_st_dt) == "Easter Island Standard Time"
    assert target_win_tzinfo.datetime.tzinfo.tzname(local_east_dt_dt) == "Easter Island Daylight Time"

    utc_east_st_dt = target_win_tzinfo.datetime.to_utc(local_east_st_dt)
    utc_east_dt_dt = target_win_tzinfo.datetime.to_utc(local_east_dt_dt)
    assert utc_east_st_dt == target_win_tzinfo.datetime.to_utc(naive_dt_may)
    assert utc_east_dt_dt == target_win_tzinfo.datetime.to_utc(naive_dt_march)
    assert utc_east_st_dt.tzinfo == datetime.timezone.utc
    assert utc_east_dt_dt.tzinfo == datetime.timezone.utc
    assert str(utc_east_st_dt) == "2019-05-04 18:00:00+00:00"
    assert str(utc_east_dt_dt) == "2019-03-04 17:00:00+00:00"

    eu_tzinfo = target_win_tzinfo.datetime.tz("W. Europe Standard Time")
    local_eu_st_dt = naive_dt_march.replace(tzinfo=eu_tzinfo)
    local_eu_dt_dt = naive_dt_may.replace(tzinfo=eu_tzinfo)
    assert str(local_eu_st_dt) == "2019-03-04 12:00:00+01:00"
    assert str(local_eu_dt_dt) == "2019-05-04 12:00:00+02:00"
    assert eu_tzinfo.tzname(local_eu_st_dt) == "W. Europe Standard Time"
    assert eu_tzinfo.tzname(local_eu_dt_dt) == "W. Europe Daylight Time"

    utc_eu_st_dt = local_eu_st_dt.astimezone(datetime.timezone.utc)
    utc_eu_dt_dt = local_eu_dt_dt.astimezone(datetime.timezone.utc)
    assert str(utc_eu_st_dt) == "2019-03-04 11:00:00+00:00"
    assert str(utc_eu_dt_dt) == "2019-05-04 10:00:00+00:00"

    # Test the switch moment to DST
    assert not eu_tzinfo.is_dst(datetime.datetime(2022, 3, 27, 2, 0, 0, tzinfo=eu_tzinfo))
    assert eu_tzinfo.is_dst(datetime.datetime(2022, 3, 27, 3, 0, 0, tzinfo=eu_tzinfo))

    # Test utc tzinfo is_dst
    utc_tzinfo = target_win_tzinfo.datetime.tz("UTC")
    assert not utc_tzinfo.is_dst(datetime.datetime(2022, 3, 27, 2, 0, 0, tzinfo=utc_tzinfo))


def test_windows_timezone_legacy(target_win_tzinfo_legacy: Target) -> None:
    # Older Windows version (prior to Windows 7) use localized time zone ids like "Paaseiland"
    target_win_tzinfo_legacy.add_plugin(WindowsDateTimePlugin)
    assert target_win_tzinfo_legacy.datetime.tzinfo.display == "(UTC-06:00) Easter Island"


def test_windows_datetime_foreign(target_win_users: Target, hive_hku: RegistryHive, hive_hklm: RegistryHive) -> None:
    # add keyboards
    preload_key_name = "Keyboard Layout\\Preload"
    preload_key = VirtualKey(hive_hku, preload_key_name)
    preload_key.add_value("1", VirtualValue(hive_hku, "1", "00000413"))
    preload_key.add_value("2", VirtualValue(hive_hku, "2", "00000409"))
    hive_hku.map_key(preload_key_name, preload_key)

    # add keyboard languages
    doskeybcodes_key_name = "SYSTEM\\ControlSet001\\Control\\Keyboard Layout\\DosKeybCodes"
    doskeybcodes_key = VirtualKey(hive_hklm, doskeybcodes_key_name)
    doskeybcodes_key.add_value("00000413", VirtualValue(hive_hklm, "00000413", "nl"))
    doskeybcodes_key.add_value("00000409", VirtualValue(hive_hklm, "00000409", "us"))
    hive_hklm.map_key(doskeybcodes_key_name, doskeybcodes_key)

    # add PST timezone
    tz_data_path = "Software\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones\\Pacific Standard Time"
    tz_data = VirtualKey(hive_hklm, tz_data_path)

    tz_data.add_value("Display", "(UTC-08:00) Some Localized Translation")
    tz_data.add_value("Dlt", "Some Localized Translation DLT")
    tz_data.add_value("Std", "Some Localized Translation STD")

    tz_data.add_value("MUI_Display", "@tzres.dll,-210")
    tz_data.add_value("MUI_Dlt", "@tzres.dll,-211")
    tz_data.add_value("MUI_Std", "@tzres.dll,-212")

    tzi = bytes.fromhex("e001000000000000c4ffffff00000b0000000100020000000000000000000300000002000200000000000000")
    tz_data.add_value("TZI", tzi)
    hive_hklm.map_key(tz_data_path, tz_data)

    # set configured language
    userprofile_key_name = "Control Panel\\International\\User Profile"
    userprofile_key = VirtualKey(hive_hku, userprofile_key_name)
    subkey = VirtualKey(hive_hku, "en-US")
    subkey.add_value("CachedLanguageName", "@Winlangdb.dll,-1337")
    userprofile_key.add_subkey("en-US", subkey)
    hive_hku.map_key(userprofile_key_name, userprofile_key)

    # set configured timezone
    timezoneinformation_key_name = "SYSTEM\\ControlSet001\\Control\\TimeZoneInformation"
    timezoneinformation_key = VirtualKey(hive_hklm, timezoneinformation_key_name)
    timezoneinformation_key.add_value("TimeZoneKeyName", "Pacific Standard Time")
    hive_hklm.map_key(timezoneinformation_key_name, timezoneinformation_key)

    target_win_users.add_plugin(WindowsDateTimePlugin)
    assert target_win_users.datetime.tzinfo.display == "(UTC-08:00) Pacific Time (US & Canada)"
    assert target_win_users.datetime.tzinfo.std_name == "Pacific Standard Time"
    assert target_win_users.datetime.tzinfo.dlt_name == "Pacific Daylight Time"

    target_win_users.add_plugin(WindowsLocalePlugin)
    assert target_win_users.timezone == "America/Los_Angeles"


def test_parse_systemtime_transition() -> None:
    # Test behaviour where not all week days are defined in every week for that month
    systemtime = c_tz._SYSTEMTIME(wDay=5, wMonth=10)
    output = parse_systemtime_transition(systemtime, 2025)
    assert output == datetime.datetime(2025, 10, 26, tzinfo=None)  # noqa

    # Test behaviour where a weekday of the month occurs 5 times during that month
    systemtime = c_tz._SYSTEMTIME(wDay=5, wMonth=10, wDayOfWeek=4)
    output = parse_systemtime_transition(systemtime, 2025)
    assert output == datetime.datetime(2025, 10, 30, tzinfo=None)  # noqa

    # wDay in this case should only go between 1-5, so this should crash
    systemtime = c_tz._SYSTEMTIME(wDay=6)
    with pytest.raises(ValueError, match="systemtime.wDay should be between 1 and 5"):
        parse_systemtime_transition(systemtime, 2025)

    systemtime = c_tz._SYSTEMTIME(wDay=0)
    with pytest.raises(ValueError, match="systemtime.wDay should be between 1 and 5"):
        parse_systemtime_transition(systemtime, 2025)



from __future__ import annotations

from datetime import datetime, timezone
from enum import IntEnum
from typing import TYPE_CHECKING

from dissect.util.ts import wintimestamp

from dissect.target.exceptions import RegistryKeyNotFoundError, UnsupportedPluginError
from dissect.target.helpers import regutil
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

AMCACHE_FILE_KEYS = {
    "0": "product_name",
    "1": "company_name",
    "2": "file_version_number",
    "3": "language_code",
    "4": "switch_back_context",
    "5": "file_version_string",
    "6": "file_size",
    "7": "pe_size_of_image",
    "8": "pe_header_hash",
    "9": "pe_header_checksum",
    "c": "file_description",
    "f": "link_timestamp",
    "11": "last_modified_timestamp",
    "12": "created_timestamp",
    "15": "full_path",
    "17": "last_modified_store_timestamp",
    "100": "program_id",
    "101": "sha1",
}

AMCACHE_PROGRAM_KEYS = {
    "a": "InstallDate",
    "0": "Name",
    "1": "Version",
    "2": "Publisher",
    "3": "LanguageCode",
    "6": "EntryType",
    "7": "UninstallKey",
    "d": "FilePaths",
    "f": "ProductCode",
    "10": "PackageCode",
    "11": "MisPackageCode",
    "12": "MisPackageCode2",
    "Files": "Files",
}

ShortcutAppcompatRecord = TargetRecordDescriptor(
    "windows/appcompat/ApplicationShortcut",
    [
        ("datetime", "mtime_regf"),
        ("path", "path"),
    ],
)

FileAppcompatRecord = TargetRecordDescriptor(
    "windows/appcompat/file",
    [
        ("datetime", "last_modified_timestamp"),
        ("datetime", "last_modified_store_timestamp"),
        ("datetime", "link_timestamp"),
        ("datetime", "created_timestamp"),
        ("datetime", "mtime_regf"),
        ("varint", "reference"),
        ("path", "path"),
        ("string", "language_code"),
        ("digest", "digest"),
        ("string", "program_id"),
        ("string", "pe_header_checksum"),
        ("string", "pe_size_of_image"),
        ("wstring", "product_name"),
        ("wstring", "company_name"),
        ("filesize", "file_size"),
    ],
)

ProgramsAppcompatRecord = TargetRecordDescriptor(
    "windows/appcompat/programs",
    [
        ("datetime", "install_date"),
        ("datetime", "mtime_regf"),
        ("wstring", "name"),
        ("string", "version"),
        ("wstring", "publisher"),
        ("string", "language_code"),
        ("string", "entry_type"),
        ("string", "uninstall_key"),
        ("path", "path"),
        ("string", "product_code"),
        ("string", "package_code"),
        ("string", "msi_package_code"),
        ("string", "msi_package_code2"),
    ],
)

ApplicationAppcompatRecord = TargetRecordDescriptor(
    "windows/appcompat/InventoryApplication",
    [
        ("datetime", "install_date"),
        ("datetime", "install_date_arp_last_modified"),
        ("datetime[]", "install_date_from_link_file"),
        ("datetime", "mtime_regf"),
        ("string", "language_code"),
        ("string", "msi_package_code"),
        ("string", "msi_product_code"),
        ("string", "name"),
        ("string", "package_full_name"),
        ("string", "type"),
        ("string", "manifest_path"),
        ("string", "os_version_at_install_time"),
        ("string", "program_id"),
        ("string", "program_instance_id"),
        ("string", "publisher"),
        ("string", "registry_key_path"),
        ("path", "root_dir_path"),
        ("string", "source"),
        ("string", "uninstall_string"),
    ],
)


ApplicationFileAppcompatRecord = TargetRecordDescriptor(
    "windows/appcompat/InventoryApplicationFile",
    [
        ("datetime", "mtime_regf"),
        ("string", "program_id"),
        ("digest", "digest"),
        ("path", "path"),
        ("string", "hash_path"),
        ("wstring", "name"),
        ("wstring", "publisher"),
        ("string", "version"),
        ("string", "bin_file_version"),
        ("wstring", "product_name"),
        ("string", "product_version"),
        ("datetime", "link_date"),
        ("string", "bin_product_version"),
        ("filesize", "size"),
        ("string", "language"),
        ("varint", "is_pefile"),
        ("varint", "is_oscomponent"),
    ],
)

BinaryAppcompatRecord = TargetRecordDescriptor(
    "windows/appcompat/InventoryDriverBinary",
    [
        ("datetime", "mtime_regf"),
        ("path", "driver_name"),
        ("path", "inf"),
        ("string", "driver_version"),
        ("wstring", "product"),
        ("string", "product_version"),
        ("string", "wdf_version"),
        ("wstring", "driver_company"),
        ("string", "driver_package_strong_name"),
        ("string", "service"),
        ("string", "driver_signed"),
        ("varint", "driver_is_kernel_mode"),
        ("datetime", "last_write_time"),
        ("datetime", "driver_timestamp"),
        ("filesize", "image_size"),
    ],
)


ContainerAppcompatRecord = TargetRecordDescriptor(
    "windows/appcompat/DeviceContainer",
    [
        ("datetime", "mtime_regf"),
        ("string", "categories"),
        ("string", "discovery_method"),
        ("string", "friendly_name"),
        ("string", "icon"),
        ("varint", "is_active"),
        ("varint", "is_connected"),
        ("varint", "is_machine_container"),
        ("varint", "is_networked"),
        ("varint", "is_paired"),
        ("string", "manufacturer"),
        ("string", "model_id"),
        ("string", "model_name"),
        ("string", "model_number"),
        ("string", "primary_category"),
        ("string", "state"),
    ],
)

PcaAppLaunchAppcompatRecord = TargetRecordDescriptor(
    "windows/appcompat/pca/AppLaunch",
    [
        ("datetime", "ts"),
        ("path", "path"),
        ("path", "source"),
    ],
)

PcaGeneralAppcompatRecord = TargetRecordDescriptor(
    "windows/appcompat/pca/General",
    [
        ("datetime", "ts"),
        ("path", "path"),
        ("string", "type"),
        ("string", "name"),
        ("string", "copyright"),
        ("string", "version"),
        ("string", "program_id"),
        ("string", "exit_message"),
        ("path", "source"),
    ],
)


class PcaGeneralDbType(IntEnum):
    """``PcaGeneralDb`` type enum.

    Resource:
        - https://www.sygnia.co/blog/new-windows-11-pca-artifact/
    """

    INSTALLER_FAILED = 0
    DRIVER_BLOCKED = 1
    ABNORMAL_PROCESS_EXIT = 2
    PCA_RESOLVE_CALLED = 3
    UNKNOWN = 4


class AmcachePluginOldMixin:
    __namespace__ = "amcache"

    def _replace_indices_with_fields(self, mapping: dict[str, str], record: regutil.RegistryKey) -> dict[str, str]:
        record_data = {v.name: v.value for v in record.values()}
        result = {}
        for key, value in record_data.items():
            field = mapping.get(key, key)
            result[field] = value
        return result

    def parse_file(self) -> Iterator[FileAppcompatRecord]:
        key = "Root\\File"

        for entry in self.read_key_subkeys(key):
            for subkey in entry.subkeys():
                subkey_data = self._replace_indices_with_fields(AMCACHE_FILE_KEYS, subkey)

                yield FileAppcompatRecord(
                    last_modified_store_timestamp=parse_win_timestamp(subkey_data.get("last_modified_store_timestamp")),
                    last_modified_timestamp=parse_win_timestamp(subkey_data.get("last_modified_timestamp")),
                    link_timestamp=subkey_data.get("link_timestamp"),
                    created_timestamp=parse_win_timestamp(subkey_data.get("created_timestamp")),
                    mtime_regf=subkey.timestamp,
                    reference=int(subkey.name, 16),
                    path=self.target.fs.path(subkey_data["full_path"]) if subkey_data.get("full_path") else None,
                    language_code=subkey_data.get("language_code"),
                    digest=(None, subkey_data["sha1"][-40:] if subkey_data.get("sha1") else None, None),
                    program_id=subkey_data.get("program_id"),
                    pe_header_checksum=subkey_data.get("pe_header_checksum"),
                    pe_size_of_image=subkey_data.get("pe_size_of_image"),
                    product_name=subkey_data.get("product_name"),
                    company_name=subkey_data.get("company_name"),
                    file_size=subkey_data.get("file_size"),
                    _target=self.target,
                )

    def parse_programs(self) -> Iterator[ProgramsAppcompatRecord]:
        key = "Root\\Programs"

        for entry in self.read_key_subkeys(key):
            entry_data = self._replace_indices_with_fields(AMCACHE_PROGRAM_KEYS, entry)

            yield ProgramsAppcompatRecord(
                mtime_regf=entry.timestamp,
                install_date=parse_win_timestamp(entry_data.get("InstallDate")),
                name=entry_data.get("Name"),
                version=entry_data.get("Version"),
                publisher=entry_data.get("Publisher"),
                language_code=entry_data.get("LanguageCode"),
                entry_type=entry_data.get("EntryType"),
                uninstall_key=entry_data.get("UninstallKey"),
                product_code=entry_data.get("ProductCode"),
                package_code=entry_data.get("PackageCode"),
                msi_package_code=entry_data.get("MsiPackageCode"),
                msi_package_code2=entry_data.get("MsiPackageCode2"),
                _target=self.target,
            )

            if "FilePaths" in entry_data:
                for file_path_entry in entry_data["FilePaths"]:
                    yield ProgramsAppcompatRecord(
                        mtime_regf=entry.timestamp,
                        install_date=parse_win_timestamp(entry_data.get("InstallDate")),
                        name=entry_data.get("Name"),
                        version=entry_data.get("Version"),
                        publisher=entry_data.get("Publisher"),
                        language_code=entry_data.get("LanguageCode"),
                        entry_type=entry_data.get("EntryType"),
                        uninstall_key=entry_data.get("UninstallKey"),
                        path=self.target.fs.path(file_path_entry),
                        product_code=entry_data.get("ProductCode"),
                        package_code=entry_data.get("PackageCode"),
                        msi_package_code=entry_data.get("MsiPackageCode"),
                        msi_package_code2=entry_data.get("MsiPackageCode2"),
                        _target=self.target,
                    )

            if "Files" in entry_data:
                for file_entry in entry_data["Files"]:
                    yield ProgramsAppcompatRecord(
                        mtime_regf=entry.timestamp,
                        install_date=parse_win_timestamp(entry_data.get("InstallDate")),
                        name=entry_data.get("Name"),
                        version=entry_data.get("Version"),
                        publisher=entry_data.get("Publisher"),
                        language_code=entry_data.get("LanguageCode"),
                        entry_type=entry_data.get("EntryType"),
                        uninstall_key=entry_data.get("UninstallKey"),
                        path=self.target.fs.path(file_entry),
                        product_code=entry_data.get("ProductCode"),
                        package_code=entry_data.get("PackageCode"),
                        msi_package_code=entry_data.get("MsiPackageCode"),
                        msi_package_code2=entry_data.get("MsiPackageCode2"),
                        _target=self.target,
                    )

    @export(record=ProgramsAppcompatRecord)
    def programs(self) -> Iterator[ProgramsAppcompatRecord]:
        """Return Programs records from Amcache hive."""
        if self.amcache:
            yield from self.parse_programs()

    @export(record=FileAppcompatRecord)
    def files(self) -> Iterator[FileAppcompatRecord]:
        """Return File records from Amcache hive."""
        if self.amcache:
            yield from self.parse_file()


class AmcachePlugin(AmcachePluginOldMixin, Plugin):
    """Appcompat plugin for amcache.hve.

    Supported registry keys:

        for old version of Amcache:
        * File
        * Programs

        for new version of Amcache:
        • InventoryDriverBinary
        • InventoryDeviceContainer
        • InventoryApplication
        • InventoryApplicationFile
        * InventoryApplicationShortcut

    References:
        - https://binaryforay.blogspot.com/2015/04/appcompatcache-changes-in-windows-10.html
        - https://cyber.gouv.fr/sites/default/files/2019/01/anssi-coriin_2019-analysis_amcache.pdf
        - https://aboutdfir.com/new-windows-11-pro-22h2-evidence-of-execution-artifact/
    """

    __namespace__ = "amcache"

    def __init__(self, target: Target):
        super().__init__(target)
        self.amcache = regutil.HiveCollection()
        self.amcache_applaunch = False

        fpath = self.target.resolve("%windir%/appcompat/programs/amcache.hve")
        if fpath.exists():
            self.amcache.add(regutil.RegfHive(fpath))

    def check_compatible(self) -> None:
        if not self.amcache and not next(self.target.resolve("%windir%/appcompat/pca").glob("Pca*.txt"), None):
            raise UnsupportedPluginError("Could not load amcache.hve or find AppLaunchDic")

    def read_key_subkeys(self, key: str) -> Iterator[regutil.RegistryKey]:
        try:
            yield from self.amcache.key(key).subkeys()
        except RegistryKeyNotFoundError:
            self.target.log.warning('Could not find registry key "%s"', key)

    def parse_inventory_application(self) -> Iterator[ApplicationAppcompatRecord]:
        """Parse Root\\InventoryApplication registry key subkeys.

        References:
            - https://docs.microsoft.com/en-us/windows/privacy/required-windows-diagnostic-data-events-and-fields-2004#microsoftwindowsinventorycoreinventoryapplicationadd
        """
        key = "Root\\InventoryApplication"

        for entry in self.read_key_subkeys(key):
            entry_data = {v.name: v.value for v in entry.values()}

            # Example:
            # {
            #      "BundleManifestPath": "",
            #      "HiddenArp": 0,
            #      "InboxModernApp": 0,
            #      "InstallDate": "07/21/2021 10:50:31",
            #      "Language": 65535,
            #      "ManifestPath": "",
            #      "MsiPackageCode": "",
            #      "MsiProductCode": "",
            #      "Name": "Shadow Tactics: Blades of the Shogun Demo",
            #      "OSVersionAtInstallTime": "10.0.0.19043",
            #      "PackageFullName": "",
            #      "ProgramId": "000002f1bc1777c6cb2bc117ab9cbd6d92780000ffff",
            #      "ProgramInstanceId": "0000a90d11947726f9702ca9ca3c76dc7631e35e861f",
            #      "Publisher": "Mimimi Games",
            #      "RegistryKeyPath":
            #           "HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\Steam App 547490",
            #      "RootDirPath":
            #           "C:\\Program Files (x86)\\Steam\\steamapps\\common\\Shadow Tactics Blades of the Shogun Demo",
            #      "SentDetailedInv": 0,
            #      "Source": "Steam",
            #      "StoreAppType": "",
            #      "Type": "Application",
            #      "UninstallString": "\"C:\\Program Files (x86)\\Steam\\steam.exe\" steam://uninstall/547490",
            #      "Version": ""
            # }

            install_date_arp_last_modified = (
                entry_data["InstallDateArpLastModified"][0] if entry_data.get("InstallDateArpLastModified") else None
            )

            yield ApplicationAppcompatRecord(
                mtime_regf=entry.timestamp,
                install_date=parse_win_datetime(entry_data.get("InstallDate")),
                install_date_arp_last_modified=parse_win_datetime(install_date_arp_last_modified),
                install_date_from_link_file=[
                    parse_win_datetime(dt) for dt in entry_data.get("InstallDateFromLinkFile", [])
                ],
                language_code=entry_data.get("Language"),
                manifest_path=entry_data.get("ManifestPath"),
                msi_package_code=entry_data.get("MsiPackageCode"),
                msi_product_code=entry_data.get("MsiProductCode"),
                name=entry_data.get("Name"),
                os_version_at_install_time=entry_data.get("OSVersionAtInstallTime"),
                package_full_name=entry_data.get("PackageFullName"),
                program_id=entry_data.get("ProgramId"),
                program_instance_id=entry_data.get("ProgramInstanceId"),
                publisher=entry_data.get("Publisher"),
                registry_key_path=entry_data.get("RegistryKeyPath"),
                root_dir_path=self.target.fs.path(entry_data.get("RootDirPath")),
                source=entry_data.get("Source"),
                uninstall_string=entry_data.get("UninstallString"),
                type=entry_data.get("Type"),
                _target=self.target,
            )

    def parse_inventory_application_file(self) -> Iterator[ApplicationFileAppcompatRecord]:
        """Parse Root\\InventoryApplicationFile registry key subkeys.

        References:
            - https://docs.microsoft.com/en-us/windows/privacy/required-windows-diagnostic-data-events-and-fields-2004#microsoftwindowsinventorycoreinventoryapplicationadd
        """
        key = "Root\\InventoryApplicationFile"

        for entry in self.read_key_subkeys(key):
            entry_data = {v.name: v.value for v in entry.values()}

            # Example:
            # {
            #     "AppxPackageFullName": "Microsoft.Microsoft3DViewer_7.2105.4012.0_x64__8wekyb3d8bbwe",
            #     "AppxPackageRelativeId": "Microsoft.Microsoft3DViewer",
            #     "BinFileVersion": "7.2105.4012.0",
            #     "BinProductVersion": "7.2105.4012.0",
            #     "BinaryType": "pe64_amd64",
            #     "FileId": "00008e01cdeb9a1c23cee421a647f29c45f67623be97",
            #     "Language": 0,
            #     "LinkDate": "05/04/2021 17:43:39",
            #     "LongPathHash": "3dviewer.exe|40f275349895ac70",
            #     "LowerCaseLongPath": "c:\\program files\\windowsapps\\microsoft.0_x64__8wekyb3d8bbwe\\3dviewer.exe",
            #     "Name": "3DViewer.exe",
            #     "OriginalFileName": "3dviewer.exe",
            #     "ProductName": "view 3d",
            #     "ProductVersion": "7.2105.4012.0",
            #     "ProgramId": "0000df892556c2f7a6b7fa69f7009b5c08cb00000904",
            #     "Publisher": "microsoft corporation",
            #     "Size": 19456,
            #     "Usn": 32259776,
            #     "Version": "7.2105.4012.0"
            # }

            sha1_digest = entry_data.get("FileId")
            # The FileId, if it exists, is always prefixed with 4 zeroes (0000)
            # and sometimes the FileId is an empty string.
            sha1_digest = sha1_digest[-40:] if sha1_digest else None

            yield ApplicationFileAppcompatRecord(
                mtime_regf=entry.timestamp,
                program_id=entry_data.get("ProgramId"),
                digest=(None, sha1_digest, None),
                path=self.target.fs.path(entry_data.get("LowerCaseLongPath")),
                link_date=parse_win_datetime(entry_data.get("LinkDate")),
                hash_path=entry_data.get("LongPathHash"),
                name=entry_data.get("Name"),
                publisher=entry_data.get("Publisher"),
                version=entry_data.get("Version"),
                bin_file_version=entry_data.get("BinFileVersion"),
                product_name=entry_data.get("ProductName"),
                product_version=entry_data.get("ProductVersion"),
                bin_product_version=entry_data.get("BinProductVersion"),
                size=entry_data.get("Size"),
                language=entry_data.get("Language"),
                is_pefile=entry_data.get("IsPeFile"),
                is_oscomponent=entry_data.get("IsOsComponent"),
                _target=self.target,
            )

    def parse_inventory_driver_binary(self) -> Iterator[BinaryAppcompatRecord]:
        key = "Root\\InventoryDriverBinary"

        for entry in self.read_key_subkeys(key):
            entry_data = {v.name: v.value for v in entry.values()}

            yield BinaryAppcompatRecord(
                mtime_regf=entry.timestamp,
                driver_name=self.target.fs.path(entry_data.get("DriverName")),
                inf=self.target.fs.path(entry_data.get("Inf")),
                driver_version=entry_data.get("DriverVersion"),
                product=entry_data.get("Product"),
                product_version=entry_data.get("ProductVersion"),
                wdf_version=entry_data.get("WdfVersion"),
                driver_company=entry_data.get("DriverCompany"),
                driver_package_strong_name=entry_data.get("DriverPackageStrongName"),
                service=entry_data.get("Service"),
                driver_signed=entry_data.get("DriverSigned"),
                driver_is_kernel_mode=entry_data.get("DriverIsKernelMode"),
                last_write_time=parse_win_datetime(entry_data.get("DriverLastWriteTime")),
                driver_timestamp=parse_win_timestamp(entry_data.get("DriverTimestamp")),
                image_size=entry_data.get("ImageSize"),
                _target=self.target,
            )

    def parse_inventory_application_shortcut(self) -> Iterator[ShortcutAppcompatRecord]:
        key = "Root\\InventoryApplicationShortcut"

        for entry in self.read_key_subkeys(key):
            yield ShortcutAppcompatRecord(
                mtime_regf=entry.timestamp,
                path=self.target.fs.path(entry.value("ShortCutPath").value),
                _target=self.target,
            )

    def parse_inventory_device_container(self) -> Iterator[ContainerAppcompatRecord]:
        # https://binaryforay.blogspot.com/2017/10/amcache-still-rules-everything-around.html
        key = "Root\\InventoryDeviceContainer"

        for entry in self.read_key_subkeys(key):
            entry_data = {v.name: v.value for v in entry.values()}
            yield ContainerAppcompatRecord(
                mtime_regf=entry.timestamp,
                categories=entry_data.get("Categories"),
                discovery_method=entry_data.get("DiscoveryMethod"),
                friendly_name=entry_data.get("FriendlyName"),
                icon=entry_data.get("Icon"),
                is_active=entry_data.get("IsActive"),
                is_connected=entry_data.get("IsConnected"),
                is_machine_container=entry_data.get("IsMachineContainer"),
                is_networked=entry_data.get("IsNetworked"),
                is_paired=entry_data.get("IsPaired"),
                manufacturer=entry_data.get("Manufacturer"),
                model_id=entry_data.get("ModelID"),
                model_name=entry_data.get("ModelName"),
                model_number=entry_data.get("ModelNumber"),
                primary_category=entry_data.get("PrimaryCategory"),
                state=entry_data.get("State"),
                _target=self.target,
            )

    @export(record=ApplicationAppcompatRecord)
    def applications(self) -> Iterator[ApplicationAppcompatRecord]:
        """Return InventoryApplication records from Amcache hive.

        Amcache is a registry hive that stores information about executed programs. The InventoryApplication key holds
        all application objects that are in cache.

        References:
            - https://docs.microsoft.com/en-us/windows/privacy/basic-level-windows-diagnostic-events-and-fields-1803
            - https://www.andreafortuna.org/2017/10/16/amcache-and-shimcache-in-forensic-analysis/
        """
        if self.amcache:
            yield from self.parse_inventory_application()

    @export(record=ApplicationFileAppcompatRecord)
    def application_files(self) -> Iterator[ApplicationFileAppcompatRecord]:
        """Return InventoryApplicationFile records from Amcache hive.

        Amcache is a registry hive that stores information about executed programs. The InventoryApplicationFile key
        holds the application files that are in cache.

        References:
            - https://docs.microsoft.com/en-us/windows/privacy/basic-level-windows-diagnostic-events-and-fields-1803
            - https://www.andreafortuna.org/2017/10/16/amcache-and-shimcache-in-forensic-analysis/
        """
        if self.amcache:
            yield from self.parse_inventory_application_file()

    @export(record=BinaryAppcompatRecord)
    def drivers(self) -> Iterator[BinaryAppcompatRecord]:
        """Return InventoryDriverBinary records from Amcache hive.

        Amcache is a registry hive that stores information about executed programs. The InventoryDriverBinary key holds
        the driver binaries that are in cache.

        References:
            - https://binaryforay.blogspot.com/2017/10/amcache-still-rules-everything-around.html
            - https://docs.microsoft.com/en-us/windows/privacy/basic-level-windows-diagnostic-events-and-fields-1803
            - https://www.andreafortuna.org/2017/10/16/amcache-and-shimcache-in-forensic-analysis/
        """
        if self.amcache:
            yield from self.parse_inventory_driver_binary()

    @export(record=ShortcutAppcompatRecord)
    def shortcuts(self) -> Iterator[ShortcutAppcompatRecord]:
        """Return InventoryApplicationShortcut records from Amcache hive.

        Amcache is a registry hive that stores information about executed programs. The InventoryApplicationShortcut
        field holds the shortcuts that are in cache. The key values contain information about the target of the lnk
        file.

        References:
            - https://binaryforay.blogspot.com/2017/10/amcache-still-rules-everything-around.html
            - https://docs.microsoft.com/en-us/windows/privacy/basic-level-windows-diagnostic-events-and-fields-1803
            - https://www.andreafortuna.org/2017/10/16/amcache-and-shimcache-in-forensic-analysis/
        """
        if self.amcache:
            yield from self.parse_inventory_application_shortcut()

    @export(record=ContainerAppcompatRecord)
    def device_containers(self) -> Iterator[ContainerAppcompatRecord]:
        """Return InventoryDeviceContainer records from Amcache hive.

        Amcache is a registry hive that stores information about executed programs. The InventoryDeviceContainer key
        holds the device containers that are in cache. Example devices are bluetooth, printers, audio, etc.

        References:
            - https://binaryforay.blogspot.com/2017/10/amcache-still-rules-everything-around.html
            - https://docs.microsoft.com/en-us/windows/privacy/basic-level-windows-diagnostic-events-and-fields-1803
            - https://www.andreafortuna.org/2017/10/16/amcache-and-shimcache-in-forensic-analysis/
        """
        if self.amcache:
            yield from self.parse_inventory_device_container()

    @export(record=PcaAppLaunchAppcompatRecord)
    def applaunches(self) -> Iterator[PcaAppLaunchAppcompatRecord]:
        """Return PcaAppLaunchAppcompatRecord records from Amcache PCA AppLaunch files (Windows 11 22H2 or later).

        References:
            - https://aboutdfir.com/new-windows-11-pro-22h2-evidence-of-execution-artifact/
        """

        if (path := self.target.resolve("%windir%/appcompat/pca/PcaAppLaunchDic.txt")).exists():
            for line in path.open("rt"):
                if not (line := line.strip()):
                    continue

                parts = line.split("|")
                if len(parts) != 2:
                    self.target.log.warning("Invalid line in PcaAppLaunchDic.txt, ignoring: %s", line)
                    continue

                app_path, ts = parts

                yield PcaAppLaunchAppcompatRecord(
                    ts=datetime.strptime(ts, "%Y-%m-%d %H:%M:%S.%f").replace(tzinfo=timezone.utc),
                    path=self.target.fs.path(app_path),
                    source=path,
                    _target=self.target,
                )

    @export(record=PcaGeneralAppcompatRecord)
    def general(self) -> Iterator[PcaGeneralAppcompatRecord]:
        """Return PcaGeneralAppcompatRecord records from Amcache PCA General files (Windows 11 22H2 or later).

        References:
            - https://aboutdfir.com/new-windows-11-pro-22h2-evidence-of-execution-artifact/
            - https://www.sygnia.co/blog/new-windows-11-pca-artifact/
        """

        for path in self.target.resolve("%windir%/appcompat/pca").glob("PcaGeneralDb*.txt"):
            for line in path.open("rt", encoding="utf-16-le"):
                if not (line := line.strip()):
                    continue

                parts = line.split("|")
                if len(parts) != 8:
                    self.target.log.warning("Invalid line in %s, ignoring: %s", path.name, line)
                    continue

                ts, type_, app_path, name, copyright, version, program_id, exit_message = parts

                yield PcaGeneralAppcompatRecord(
                    ts=datetime.strptime(ts, "%Y-%m-%d %H:%M:%S.%f").replace(tzinfo=timezone.utc),
                    path=self.target.resolve(app_path),
                    type=PcaGeneralDbType(int(type_)).name,
                    name=name,
                    copyright=copyright,
                    version=version,
                    program_id=program_id,
                    exit_message=exit_message,
                    source=path,
                    _target=self.target,
                )


def parse_win_datetime(value: str) -> datetime | None:
    if value:
        return datetime.strptime(value, "%m/%d/%Y %H:%M:%S").replace(tzinfo=timezone.utc)
    return None


def parse_win_timestamp(value: str) -> datetime | None:
    if value:
        return wintimestamp(value)
    return None



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.locale import WindowsLocalePlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_locale_plugin_windows(target_win_users: Target, hive_hku: VirtualHive, hive_hklm: VirtualHive) -> None:
    # Hive structure originates from a Windows Server 2019 installation.

    # language dict
    doskeybcodes_key_name = "SYSTEM\\ControlSet001\\Control\\Keyboard Layout\\DosKeybCodes"
    doskeybcodes_key = VirtualKey(hive_hklm, doskeybcodes_key_name)
    doskeybcodes_key.add_value("00000413", VirtualValue(hive_hku, "00000413", "nl"))
    doskeybcodes_key.add_value("00000409", VirtualValue(hive_hku, "00000409", "us"))
    hive_hklm.map_key(doskeybcodes_key_name, doskeybcodes_key)

    # installed keyboards
    preload_key_name = "Keyboard Layout\\Preload"
    preload_key = VirtualKey(hive_hku, preload_key_name)
    preload_key.add_value("1", VirtualValue(hive_hku, "1", "00000413"))
    preload_key.add_value("2", VirtualValue(hive_hku, "2", "00000409"))
    hive_hku.map_key(preload_key_name, preload_key)

    # installed languages
    userprofile_key_name = "Control Panel\\International\\User Profile"
    userprofile_key = VirtualKey(hive_hku, userprofile_key_name)
    subkey = VirtualKey(hive_hku, "en-US")
    subkey.add_value("CachedLanguageName", "@Winlangdb.dll,-1337")
    userprofile_key.add_subkey("en-US", subkey)
    hive_hku.map_key(userprofile_key_name, userprofile_key)

    # configured timezone
    timezoneinformation_key_name = "SYSTEM\\ControlSet001\\Control\\TimeZoneInformation"
    timezoneinformation_key = VirtualKey(hive_hklm, timezoneinformation_key_name)
    timezoneinformation_key.add_value("TimeZoneKeyName", "Pacific Standard Time")
    hive_hklm.map_key(timezoneinformation_key_name, timezoneinformation_key)

    # timezone info
    tz_data_path = "Software\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones\\Pacific Standard Time"
    tz_data = VirtualKey(hive_hklm, tz_data_path)
    tz_data.add_value("Display", "(UTC-08:00) Pacific Time (US & Canada)")
    tz_data.add_value("Dlt", "Pacific Summer Time")
    tz_data.add_value("Std", "Pacific Standard Time")
    tzi = bytes.fromhex("e001000000000000c4ffffff00000b0000000100020000000000000000000300000002000200000000000000")
    tz_data.add_value("TZI", tzi)
    hive_hklm.map_key(tz_data_path, tz_data)

    target_win_users.add_plugin(WindowsLocalePlugin)
    assert target_win_users.language == ["en_US"]
    assert target_win_users.timezone == "America/Los_Angeles"

    keyboard = list(target_win_users.keyboard())
    assert len(keyboard) == 2
    assert keyboard[0].layout == "nl"
    assert keyboard[0].language_id == "00000413"
    assert keyboard[1].layout == "us"
    assert keyboard[1].language_id == "00000409"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.hypervisor import hyperv

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target


class HyperVChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields from Hyper-V VM inventory.

    Since Windows Server 2016, Hyper-V VMs are registered in a data.vmcx file in the Hyper-V ProgramData directory.
    Before that, VMs were registered by having a .xml file in the Hyper-V ProgramData "Virtual Machines" directory.

    It is possible to put your "VM descriptor" files elsewhere. In the .vmcx format, the full path to the alternative
    location is stored. In the .xml format, a NTFS symlink reparse point is created that points to the full
    alternative path.
    """

    __type__ = "hyper-v"

    PATH = "sysvol/ProgramData/Microsoft/Windows/Hyper-V"

    def __init__(self, target: Target):
        super().__init__(target)

        hyperv_path = self.target.fs.path(self.PATH)
        self.data_vmcx = hyperv_path.joinpath("data.vmcx")
        self.vm_xml = list(hyperv_path.joinpath("Virtual Machines").glob("*.xml"))

    def check_compatible(self) -> None:
        if not self.data_vmcx.exists() and not self.vm_xml:
            raise UnsupportedPluginError("No registered VMs and no data.vmcx file found")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        if self.data_vmcx.exists():
            data = hyperv.HyperVFile(self.data_vmcx.open()).as_dict()

            if virtual_machines := data["Configurations"].get("VirtualMachines"):
                for vm_path in virtual_machines.values():
                    yield ChildTargetRecord(
                        type=self.__type__,
                        path=self.target.fs.path(vm_path),
                        _target=self.target,
                    )

        for xml_path in self.vm_xml:
            yield ChildTargetRecord(
                type=self.__type__,
                path=xml_path.resolve(),
                _target=self.target,
            )



from __future__ import annotations

import sys
from typing import TYPE_CHECKING
from unittest.mock import MagicMock

from dissect.target.exceptions import FileNotFoundError
from dissect.target.filesystem import FilesystemEntry
from dissect.target.helpers.fsutil import stat_result
from dissect.target.tools.fsutils import print_extensive_file_stat_listing

if TYPE_CHECKING:
    import pytest

    from dissect.target.target import Target


def test_target_cli_print_extensive_file_stat(target_win: Target, capsys: pytest.CaptureFixture) -> None:
    mock_stat = stat_result([0o100777, 1, 2, 3, 1337, 7331, 999, 0, 0, 0])
    mock_entry = MagicMock(spec_set=FilesystemEntry)
    mock_entry.lstat.return_value = mock_stat
    mock_entry.is_symlink.return_value = False

    print_extensive_file_stat_listing(sys.stdout, "foo", mock_entry)

    captured = capsys.readouterr()
    assert captured.out == "-rwxrwxrwx 1337 7331        999 1970-01-01T00:00:00.000000+00:00 foo\n"


def test_print_extensive_file_stat_symlink(target_win: Target, capsys: pytest.CaptureFixture) -> None:
    mock_stat = stat_result([0o120777, 1, 2, 3, 1337, 7331, 999, 0, 0, 0])
    mock_entry = MagicMock(spec_set=FilesystemEntry)
    mock_entry.lstat.return_value = mock_stat
    mock_entry.is_symlink.return_value = True
    mock_entry.readlink.return_value = "bar"

    print_extensive_file_stat_listing(sys.stdout, "foo", mock_entry)

    captured = capsys.readouterr()
    assert captured.out == "lrwxrwxrwx 1337 7331        999 1970-01-01T00:00:00.000000+00:00 foo -> bar\n"


def test_print_extensive_file_stat_fail(target_win: Target, capsys: pytest.CaptureFixture) -> None:
    mock_entry = MagicMock(spec_set=FilesystemEntry)
    mock_entry.lstat.side_effect = FileNotFoundError("ERROR")
    print_extensive_file_stat_listing(sys.stdout, "foo", mock_entry)

    captured = capsys.readouterr()
    assert captured.out == "??????????    ?    ?          ? ????-??-??T??:??:??.??????+??:?? foo\n"



from __future__ import annotations

import base64
import re
from itertools import product
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import export
from dissect.target.plugins.apps.ssh.ssh import (
    AuthorizedKeysRecord,
    KnownHostRecord,
    PrivateKeyRecord,
    PublicKeyRecord,
    SSHPlugin,
    SSHPrivateKey,
    calculate_fingerprints,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target


def find_sshd_directory(target: Target) -> TargetPath:
    SSHD_DIRECTORIES = ["/sysvol/ProgramData/ssh", "/etc/ssh"]

    for sshd in SSHD_DIRECTORIES:
        if (target_path := target.fs.path(sshd)).exists():
            return target_path

    # A default, so there is no need to check for None
    return target.fs.path("/etc/ssh/")


class OpenSSHPlugin(SSHPlugin):
    """OpenSSH plugin."""

    __namespace__ = "openssh"

    SSHD_DIRECTORIES = ("/sysvol/ProgramData/ssh", "/etc/ssh")

    def __init__(self, target: Target):
        super().__init__(target)
        self.sshd_directory = find_sshd_directory(target)

    def check_compatible(self) -> None:
        ssh_user_dirs = any(
            user_details.home_path.joinpath(".ssh").exists()
            for user_details in self.target.user_details.all_with_home()
        )
        if not ssh_user_dirs and not self.sshd_directory.exists():
            raise UnsupportedPluginError("No OpenSSH directories found")

    def ssh_directory_globs(self, glob_user: str, glob_sshd: str) -> list[tuple[str, TargetPath]]:
        for user_details in self.target.user_details.all_with_home():
            yield from product([user_details.user], user_details.home_path.glob(f".ssh/{glob_user}"))

        yield from product([None], self.sshd_directory.glob(glob_sshd))

    @export(record=AuthorizedKeysRecord)
    def authorized_keys(self) -> Iterator[AuthorizedKeysRecord]:
        """Yields the content of the authorized_keys files on a target for each user."""
        for user, authorized_keys_file in self.ssh_directory_globs("authorized_keys*", "administrator_authorized_keys"):
            for line in authorized_keys_file.open("rt"):
                line = line.strip()
                if not line or line.startswith("#"):
                    continue

                try:
                    options, keytype, public_key, comment = parse_ssh_key(line)
                except ValueError:
                    continue

                yield AuthorizedKeysRecord(
                    key_type=keytype,
                    public_key=public_key,
                    comment=comment,
                    options=options,
                    path=authorized_keys_file,
                    _target=self.target,
                    _user=user,
                )

    @export(record=KnownHostRecord)
    def known_hosts(self) -> Iterator[KnownHostRecord]:
        """Yields the content of the known_hosts files on a target for each user."""

        for user, known_hosts_file in self.ssh_directory_globs("known_hosts*", "ssh_known_hosts"):
            for line in known_hosts_file.open("rt"):
                line = line.strip()
                if not line or line.startswith("#"):
                    continue

                try:
                    marker, hostnames, keytype, public_key, comment = parse_known_host(line)
                except ValueError:
                    continue

                for hostname in hostnames:
                    yield KnownHostRecord(
                        host=hostname,
                        port=None,
                        key_type=keytype,
                        public_key=public_key,
                        comment=comment,
                        marker=marker,
                        path=known_hosts_file,
                        _target=self.target,
                        _user=user,
                    )

    @export(record=PrivateKeyRecord)
    def private_keys(self) -> Iterator[PrivateKeyRecord]:
        """Yields OpenSSH private keys on a target for each user."""

        for user, file_path in self.ssh_directory_globs("*", "*"):
            if not file_path.is_file():
                continue

            buffer = file_path.read_bytes().strip()
            if b"PRIVATE KEY-----" not in buffer:
                continue

            try:
                private_key = SSHPrivateKey(buffer)
            except ValueError as e:
                self.target.log.warning("Failed to parse SSH private key: %s", file_path)
                self.target.log.debug("", exc_info=e)
                continue

            yield PrivateKeyRecord(
                mtime_ts=file_path.stat().st_mtime,
                key_format=private_key.format,
                key_type=private_key.key_type,
                public_key=private_key.public_key,
                comment=private_key.comment,
                encrypted=private_key.is_encrypted,
                path=file_path,
                _target=self.target,
                _user=user,
            )

    @export(record=PublicKeyRecord)
    def public_keys(self) -> Iterator[PublicKeyRecord]:
        """Yields all OpenSSH public keys from all user home directories and the OpenSSH daemon directory."""

        for user, file_path in self.ssh_directory_globs("*.pub", "*.pub"):
            if not file_path.is_file():
                continue

            key_type, public_key, comment = parse_ssh_public_key_file(file_path)
            fingerprints = calculate_fingerprints(base64.b64decode(public_key))

            yield PublicKeyRecord(
                mtime_ts=file_path.stat().st_mtime,
                key_type=key_type,
                public_key=public_key,
                comment=comment,
                fingerprint=fingerprints,
                path=file_path,
                _target=self.target,
                _user=user,
            )


def parse_ssh_public_key_file(path: Path) -> tuple[str, str, str]:
    _, key_type, public_key, comment = parse_ssh_key(path.read_text().strip())
    return key_type, public_key, comment


def parse_ssh_key(key_string: str) -> tuple[str, str, str, str]:
    parts = re.findall(r'(?:[^\s,"]|"(?:\\.|[^"])*")+', key_string)

    options = None
    if not key_string.startswith(("sk-", "ssh-", "ecdsa-")):
        options = parts.pop(0)

    keytype, public_key = parts[:2]

    # We strip whitespace to make parsing a bit safer, but this means we also strip "empty" comments
    # An empty comment is technically "", so return that instead of None
    comment = " ".join(parts[2:]) if len(parts) > 2 else ""

    return options, keytype, public_key, comment


def parse_known_host(known_host_string: str) -> tuple[str, list, str, str, str]:
    parts = known_host_string.split()

    marker = None
    if parts[0].startswith("@"):
        marker = parts.pop(0)

    hostnames, keytype, public_key = parts[:3]

    # We strip whitespace to make parsing a bit safer, but this means we also strip "empty" comments
    # An empty comment is technically "", so return that instead of None
    comment = " ".join(parts[3:]) if len(parts) > 3 else ""

    return marker, hostnames.split(","), keytype, public_key, comment



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.util.ts import from_unix

from dissect.target.exceptions import FileNotFoundError, UnsupportedPluginError
from dissect.target.filesystem import FilesystemEntry, LayerFilesystemEntry
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

FilesystemRecord = TargetRecordDescriptor(
    "filesystem/entry",
    [
        ("datetime", "atime"),
        ("datetime", "mtime"),
        ("datetime", "ctime"),
        ("datetime", "btime"),
        ("varint", "ino"),
        ("path", "path"),
        ("filesize", "size"),
        ("uint32", "mode"),
        ("uint32", "uid"),
        ("uint32", "gid"),
        ("string[]", "fstypes"),
    ],
)


class WalkFSPlugin(Plugin):
    """Filesystem agnostic walkfs plugin."""

    def check_compatible(self) -> None:
        if not len(self.target.filesystems):
            raise UnsupportedPluginError("No filesystems to walk")

    @export(record=FilesystemRecord)
    @arg("--walkfs-path", default="/", help="path to recursively walk")
    def walkfs(self, walkfs_path: str = "/") -> Iterator[FilesystemRecord]:
        """Walk a target's filesystem and return all filesystem entries."""

        path = self.target.fs.path(walkfs_path)

        if not path.exists():
            self.target.log.error("No such directory: '%s'", walkfs_path)
            return

        if not path.is_dir():
            self.target.log.error("Not a directory: '%s'", walkfs_path)
            return

        for entry in self.target.fs.recurse(walkfs_path):
            try:
                yield generate_record(self.target, entry)

            except FileNotFoundError as e:  # noqa: PERF203
                self.target.log.warning("File not found: %s", entry)
                self.target.log.debug("", exc_info=e)
            except Exception as e:
                self.target.log.warning("Exception generating record for: %s", entry)
                self.target.log.debug("", exc_info=e)
                continue


def generate_record(target: Target, entry: FilesystemEntry) -> FilesystemRecord:
    """Generate a :class:`FilesystemRecord` from the given :class:`FilesystemEntry`.

    Args:
        target: :class:`Target` instance
        entry: :class:`FilesystemEntry` instance

    Returns:
        Generated :class:`FilesystemRecord` for the given :class:`FilesystemEntry`.
    """
    stat = entry.lstat()

    if isinstance(entry, LayerFilesystemEntry):
        fs_types = [sub_entry.fs.__type__ for sub_entry in entry.entries]
    else:
        fs_types = [entry.fs.__type__]

    return FilesystemRecord(
        atime=from_unix(stat.st_atime),
        mtime=from_unix(stat.st_mtime),
        ctime=from_unix(stat.st_ctime),
        btime=from_unix(stat.st_birthtime) if stat.st_birthtime else None,
        ino=stat.st_ino,
        path=entry.path,
        size=stat.st_size,
        mode=stat.st_mode,
        uid=stat.st_uid,
        gid=stat.st_gid,
        fstypes=fs_types,
        _target=target,
    )



from __future__ import annotations

import io
import math
from unittest.mock import Mock, patch

import pytest
from dissect.ntfs.attr import Attribute, StandardInformation
from dissect.ntfs.c_ntfs import c_ntfs
from dissect.ntfs.exceptions import FileNotFoundError as NtfsFileNotFoundError
from dissect.ntfs.mft import MftRecord
from dissect.ntfs.util import AttributeMap

from dissect.target.exceptions import FileNotFoundError
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.filesystems.ntfs import NtfsFilesystem, NtfsFilesystemEntry


@pytest.mark.parametrize(
    ("path", "expected_path", "expected_ads"),
    [
        ("test:data", "test", "data"),
        ("test:$hello", "test", "$hello"),
        ("test:$hello:$test", "test:$hello", "$test"),
    ],
)
def test_ads_ntfs_filesystem(path: str, expected_path: str, expected_ads: str) -> None:
    with patch("dissect.target.filesystems.ntfs.NTFS"):
        filesystem = NtfsFilesystem()
        entry = filesystem.get(path)

        assert entry.ads == expected_ads
        assert entry.path == expected_path
        filesystem.ntfs.mft.get.assert_called_once_with(expected_path, root=None)


@pytest.mark.parametrize(
    ("ads", "name", "output"),
    [
        ("ads", "", "ads"),
        ("ads", "test", "test"),
    ],
)
def test_ntfs_fileentry_open(ads: str, name: str, output: str) -> None:
    vfs = VirtualFilesystem()
    mocked_entry = Mock()
    mocked_entry.attributes = AttributeMap()
    mocked_entry.is_dir.return_value = False
    mocked_entry.is_symlink.return_value = False
    mocked_entry.is_mount_point.return_value = False
    entry = NtfsFilesystemEntry(vfs, "some/random/path", entry=mocked_entry)
    entry.ads = ads
    entry.open(name)

    mocked_entry.open.assert_called_once_with(output)


def test_ntfs_unknown_file() -> None:
    vfs = VirtualFilesystem()
    mocked_entry = Mock()
    mocked_entry.attributes = AttributeMap()
    mocked_entry.is_dir.return_value = False
    mocked_entry.is_symlink.return_value = False
    mocked_entry.is_mount_point.return_value = False
    mocked_entry.size.side_effect = [NtfsFileNotFoundError]
    entry = NtfsFilesystemEntry(vfs, "some/random/path", entry=mocked_entry)
    with pytest.raises(FileNotFoundError):
        entry.stat()


@pytest.mark.parametrize(
    ("cluster_size", "size", "resident", "expected_blks"),
    [
        (0x1000, 0x343, False, 8),
        (0x1000, 0x1001, False, 16),
        (0x1000, 0, False, 0),
        (0x1000, 0x2000, True, 0),
    ],
)
def test_stat_information(cluster_size: int, size: int, resident: bool, expected_blks: int) -> None:
    ntfs = Mock(cluster_size=cluster_size)

    entry = MftRecord()
    entry.header = c_ntfs._FILE_RECORD_SEGMENT_HEADER()
    entry.ntfs = ntfs
    entry.segment = 42

    attribute_record = c_ntfs._ATTRIBUTE_RECORD_HEADER()
    attribute_record.FormCode = 0 if resident else 1
    if resident:
        attribute_record.Form.Resident.ValueLength = size
    else:
        attribute_record.Form.Nonresident.FileSize = size
        # Needs to be a multiple of cluster_size
        attribute_record.Form.Nonresident.AllocatedLength = math.ceil(size / cluster_size) * cluster_size

    map = AttributeMap()
    map[0x10] = StandardInformation(
        io.BytesIO(
            b"\xb5\xc3S\xbb\xd1a\xd8\x01\xc1H\xedc$\x04\xdb\x01d \x0c\xb0v\xcc\xd9"
            b"\x01\xc1H\xedc$\x04\xdb\x01 \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00%\x03\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x92h8\x00\x00\x00\x00"
        )
    )
    map[0x80] = [Attribute.from_fh(io.BytesIO(attribute_record.dumps()))]

    mock_fs = Mock()
    with patch.object(entry, "attributes", map):
        fs_entry = NtfsFilesystemEntry(mock_fs, "some/path", entry)

        stat_info = fs_entry.lstat()

        assert stat_info.st_mode == 33279
        assert stat_info.st_ino == 42
        assert stat_info.st_dev == id(mock_fs)
        assert stat_info.st_nlink == 0
        assert stat_info.st_uid == 0
        assert stat_info.st_gid == 0
        assert stat_info.st_size == size

        assert stat_info.st_atime == 1726043227.939039
        assert stat_info.st_atime_ns == 1726043227939040100
        assert stat_info.st_mtime == 1726043227.939039
        assert stat_info.st_mtime_ns == 1726043227939040100
        assert stat_info.st_ctime == 1651900642.631773
        assert stat_info.st_ctime_ns == 1651900642631774900
        assert stat_info.st_birthtime == 1651900642.631773
        assert stat_info.st_birthtime_ns == 1651900642631774900

        assert stat_info.st_blksize == cluster_size
        assert stat_info.st_blocks == expected_blks



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.apps.container.podman import PodmanPlugin
from dissect.target.plugins.child.podman import PodmanChildTargetPlugin
from tests.plugins.apps.container.test_podman import target_unix_podman  # noqa: F401

if TYPE_CHECKING:
    from dissect.target import Target
    from dissect.target.filesystem import VirtualFilesystem


def test_plugins_child_podman(target_unix_podman: Target, fs_unix: VirtualFilesystem) -> None:  # noqa: F811
    """Test if we can find, parse and correctly yield child Podman targets."""

    target_unix_podman.add_plugin(PodmanPlugin)
    target_unix_podman.add_plugin(PodmanChildTargetPlugin)

    children = sorted(target_unix_podman.list_children(), key=lambda r: r.path)

    assert len(children) == 3
    assert children[0].type == "podman"

    assert sorted([c.path for c in children]) == [
        "/home/user/.local/share/containers/storage/overlay/04a40aded310ba9deffbd5b5b0120a0a4416e6083420e338e998250f1a2e2f2b",
        "/home/user/.local/share/containers/storage/overlay/5c2861226e61770d45f08a5bee9205c13c23221969d1fde7c3f4088f8aa1d46e",
        "/home/user/.local/share/containers/storage/overlay/f351129587e2bb1da9ba4f03dcd22e1c838cd4f20dcc70e6da72381d2905b913",
    ]



from __future__ import annotations

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import export
from dissect.target.plugins.os.default.locale import LocalePlugin


class FortiOSLocalePlugin(LocalePlugin):
    """FortiOS locale plugin."""

    def check_compatible(self) -> None:
        if self.target.os != "fortios":
            raise UnsupportedPluginError("FortiOS specific plugin loaded on non-FortiOS target")

    @export(property=True)
    def timezone(self) -> str | None:
        """Return configured UI/system timezone."""
        try:
            timezone_num = self.target._os._config["global-config"]["system"]["global"]["timezone"][0]
            return translate_timezone(timezone_num)
        except KeyError:
            pass

    @export(property=True)
    def language(self) -> str | None:
        """Return configured UI language."""
        LANG_MAP = {
            "english": "en_US",
            "french": "fr_FR",
            "spanish": "es_ES",
            "portuguese": "pt_PT",
            "japanese": "ja_JP",
            "trach": "zh_TW",
            "simch": "zh_CN",
            "korean": "ko_KR",
        }
        try:
            lang_str = self.target._os._config["global-config"]["system"]["global"].get("language", ["english"])[0]
            return LANG_MAP.get(lang_str, lang_str)
        except KeyError:
            pass


def translate_timezone(timezone_num: str) -> str:
    """Translate a FortiOS timezone number to IANA TZ.

    References:
        - https://<fortios>/ng/system/settings
    """

    TZ_MAP = {
        "01": "Etc/GMT+11",  # (GMT-11:00) Midway Island, Samoa
        "02": "Pacific/Honolulu",  # (GMT-10:00) Hawaii
        "03": "America/Anchorage",  # (GMT-9:00) Alaska
        "04": "America/Los_Angeles",  # (GMT-8:00) Pacific Time (US & Canada)
        "05": "America/Phoenix",  # (GMT-7:00) Arizona
        "81": "America/Chihuahua",  # (GMT-7:00) Baja California Sur, Chihuahua
        "06": "America/Denver",  # (GMT-7:00) Mountain Time (US & Canada)
        "07": "America/Guatemala",  # (GMT-6:00) Central America
        "08": "America/Chicago",  # (GMT-6:00) Central Time (US & Canada)
        "09": "America/Mexico_City",  # (GMT-6:00) Mexico City
        "10": "America/Regina",  # (GMT-6:00) Saskatchewan
        "11": "America/Bogota",  # (GMT-5:00) Bogota, Lima,Quito
        "12": "America/New_York",  # (GMT-5:00) Eastern Time (US & Canada)
        "13": "America/Indianapolis",  # (GMT-5:00) Indiana (East)
        "74": "America/Caracas",  # (GMT-4:00) Caracas
        "14": "America/Halifax",  # (GMT-4:00) Atlantic Time (Canada)
        "77": "Etc/GMT+4",  # (GMT-4:00) Georgetown
        "15": "America/La_Paz",  # (GMT-4:00) La Paz
        "87": "America/Asuncion",  # (GMT-4:00) Paraguay
        "16": "America/Santiago",  # (GMT-3:00) Santiago
        "17": "America/St_Johns",  # (GMT-3:30) Newfoundland
        "18": "America/Sao_Paulo",  # (GMT-3:00) Brasilia
        "19": "America/Buenos_Aires",  # (GMT-3:00) Buenos Aires
        "20": "America/Godthab",  # (GMT-3:00) Nuuk (Greenland)
        "75": "America/Montevideo",  # (GMT-3:00) Uruguay
        "21": "Etc/GMT+2",  # (GMT-2:00) Mid-Atlantic
        "22": "Atlantic/Azores",  # (GMT-1:00) Azores
        "23": "Atlantic/Cape_Verde",  # (GMT-1:00) Cape Verde Is.
        "24": "Atlantic/Reykjavik",  # (GMT) Monrovia
        "80": "Europe/London",  # (GMT) Greenwich Mean Time
        "79": "Africa/Casablanca",  # (GMT) Casablanca
        "25": "Etc/UTC",  # (GMT) Dublin, Edinburgh, Lisbon, London, Canary Is.
        "26": "Europe/Berlin",  # (GMT+1:00) Amsterdam, Berlin, Bern, Rome, Stockholm, Vienna
        "27": "Europe/Budapest",  # (GMT+1:00) Belgrade, Bratislava, Budapest, Ljubljana, Prague
        "28": "Europe/Paris",  # (GMT+1:00) Brussels, Copenhagen, Madrid, Paris
        "78": "Africa/Windhoek",  # (GMT+1:00) Namibia
        "29": "Europe/Warsaw",  # (GMT+1:00) Sarajevo, Skopje, Warsaw, Zagreb
        "30": "Africa/Lagos",  # (GMT+1:00) West Central Africa
        "31": "Europe/Kiev",  # (GMT+2:00) Athens, Sofia, Vilnius
        "32": "Europe/Bucharest",  # (GMT+2:00) Bucharest
        "33": "Africa/Cairo",  # (GMT+2:00) Cairo
        "34": "Africa/Johannesburg",  # (GMT+2:00) Harare, Pretoria
        "35": "Europe/Helsinki",  # (GMT+2:00) Helsinki, Riga, Tallinn
        "36": "Asia/Jerusalem",  # (GMT+2:00) Jerusalem
        "37": "Asia/Baghdad",  # (GMT+3:00) Baghdad
        "38": "Asia/Riyadh",  # (GMT+3:00) Kuwait, Riyadh
        "83": "Europe/Moscow",  # (GMT+3:00) Moscow
        "84": "Europe/Minsk",  # (GMT+3:00) Minsk
        "40": "Africa/Nairobi",  # (GMT+3:00) Nairobi
        "85": "Europe/Istanbul",  # (GMT+3:00) Istanbul
        "41": "Asia/Tehran",  # (GMT+3:30) Tehran
        "42": "Asia/Dubai",  # (GMT+4:00) Abu Dhabi, Muscat
        "43": "Asia/Baku",  # (GMT+4:00) Baku
        "39": "Europe/Volgograd",  # (GMT+3:00) St. Petersburg, Volgograd
        "44": "Asia/Kabul",  # (GMT+4:30) Kabul
        "46": "Asia/Karachi",  # (GMT+5:00) Islamabad, Karachi, Tashkent
        "47": "Asia/Calcutta",  # (GMT+5:30) Kolkata, Chennai, Mumbai, New Delhi
        "51": "Asia/Colombo",  # (GMT+5:30) Sri Jayawardenepara
    }

    return TZ_MAP.get(timezone_num, timezone_num)



from __future__ import annotations

import pathlib
import tempfile
from unittest.mock import Mock, patch

import pytest

from dissect.target.filesystems.dir import DirectoryFilesystem, DirectoryFilesystemEntry


def test_symlink_to_file(tmp_path: pathlib.Path) -> None:
    with tempfile.NamedTemporaryFile(dir=tmp_path, delete=False) as tf:
        tf.write(b"dummy")
        tf.close()

        tmpfile_path = pathlib.Path(tf.name)
        symlink_path = tmp_path.joinpath("symlink")
        symlink_path.symlink_to(f"/{tmpfile_path.name}")

        fs = DirectoryFilesystem(path=tmp_path)
        symlink_entry = fs.get("symlink")

        assert symlink_entry.is_symlink()
        assert symlink_entry.is_file()
        assert not symlink_entry.is_file(follow_symlinks=False)
        assert not symlink_entry.is_dir()
        assert symlink_entry.exists()
        assert symlink_entry.stat(follow_symlinks=False) == symlink_entry.lstat()

        assert symlink_entry.readlink() == f"/{tmpfile_path.name}"
        assert symlink_entry.readlink_ext().entry == fs.get(tmpfile_path.name).entry

        assert symlink_entry.open().read() == fs.get(tmpfile_path.name).open().read() == b"dummy"

        assert list(symlink_entry.lstat()) == list(symlink_path.lstat())
        assert list(symlink_entry.stat()) == list(tmpfile_path.lstat())


def test_symlink_to_dir(tmp_path: pathlib.Path) -> None:
    nested_path = tmp_path.joinpath("nested")
    nested_path.mkdir()
    nested_path.joinpath("file1").touch()
    nested_path.joinpath("file2").touch()

    symlink_path = tmp_path.joinpath("symlink")
    symlink_path.symlink_to("/nested")

    fs = DirectoryFilesystem(path=tmp_path)
    symlink_entry = fs.get("symlink")

    assert symlink_entry.is_symlink()
    assert not symlink_entry.is_file()
    assert symlink_entry.is_dir()
    assert not symlink_entry.is_dir(follow_symlinks=False)
    assert symlink_entry.exists()
    assert symlink_entry.stat(follow_symlinks=False) == symlink_entry.lstat()

    assert symlink_entry.readlink() == "/nested"
    assert symlink_entry.readlink_ext().entry == fs.get("/nested").entry

    assert sorted(symlink_entry.iterdir()) == ["file1", "file2"]
    assert sorted([e.entry for e in symlink_entry.scandir()], key=lambda e: e.name) == [
        fs.get("/nested/file1").entry,
        fs.get("/nested/file2").entry,
    ]


@pytest.fixture
def dirfs_entry() -> DirectoryFilesystemEntry:
    return DirectoryFilesystemEntry(Mock(), "/some/path", Mock())


def test_entry_attr(dirfs_entry: DirectoryFilesystemEntry) -> None:
    with patch("dissect.target.helpers.fsutil.fs_attrs", autospec=True) as fs_attrs:
        dirfs_entry.attr()
        fs_attrs.assert_called_with(dirfs_entry.entry, follow_symlinks=True)


def test_entry_lattr(dirfs_entry: DirectoryFilesystemEntry) -> None:
    with patch("dissect.target.helpers.fsutil.fs_attrs", autospec=True) as fs_attrs:
        dirfs_entry.lattr()
        fs_attrs.assert_called_with(dirfs_entry.entry, follow_symlinks=False)



from __future__ import annotations

from dissect.target.helpers import docs
from dissect.target.plugins.apps.webserver.iis import IISLogsPlugin


def get_nonempty_lines_set(paragraph: str) -> set[str]:
    return set(filter(None, (line.strip() for line in paragraph.splitlines())))


def test_docs_plugin_description() -> None:
    plugin_desc = docs.get_plugin_description(IISLogsPlugin)

    assert plugin_desc
    assert IISLogsPlugin.__name__ in plugin_desc

    assert get_nonempty_lines_set(IISLogsPlugin.__doc__).issubset(get_nonempty_lines_set(plugin_desc))


def test_docs_plugin_functions_desc() -> None:
    functions_short_desc = docs.get_plugin_functions_desc(IISLogsPlugin, with_docstrings=False)

    assert functions_short_desc
    desc_lines = functions_short_desc.splitlines()

    assert len(desc_lines) == 2
    assert "iis.logs" in functions_short_desc
    assert "Return contents of IIS (v7 and above) log files." in functions_short_desc
    assert "output: records" in functions_short_desc

    functions_long_desc = docs.get_plugin_functions_desc(IISLogsPlugin, with_docstrings=True)

    assert functions_long_desc

    lines_bag = get_nonempty_lines_set(functions_long_desc)

    assert "Return contents of IIS (v7 and above) log files." in lines_bag
    assert "Supported log formats: IIS, W3C." in lines_bag


def test_docs_get_func_description() -> None:
    func = IISLogsPlugin.logs
    func_desc = docs.get_func_description(func, with_docstrings=False)

    assert func_desc == "iis.logs - Return contents of IIS (v7 and above) log files. (output: records)"

    func_desc = docs.get_func_description(func, with_docstrings=True)
    lines_bag = get_nonempty_lines_set(func_desc)

    assert "Return contents of IIS (v7 and above) log files." in lines_bag
    assert "Supported log formats: IIS, W3C." in lines_bag



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.tools.yara import HAS_YARA
from dissect.target.tools.yara import main as target_yara
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.target import Target


@pytest.mark.skipif(not HAS_YARA, reason="requires python-yara")
def test_yara(target_default: Target, monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture) -> None:
    vfs = VirtualFilesystem()
    vfs.map_file_fh("test_file", BytesIO(b"hello there this is a test string!"))
    vfs.map_file_fh("/test/dir/to/test_file", BytesIO(b"this is another test string for YARA testing."))
    vfs.map_file_fh("should_not_hit", BytesIO(b"this is another file."))
    target_default.fs.mount("/", vfs)

    with patch("dissect.target.Target.open_all", return_value=[target_default]), monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "target-yara",
                "example.img",
                "--rules",
                str(absolute_path("_data/plugins/filesystem/yara/rule-dir/rule.yar")),
                "--path",
                "/",
                "--check",
                "-s",
            ],
        )
        target_yara()

        out, _ = capsys.readouterr()

        hit1 = "<filesystem/yara/match hostname=None domain=None path='/test_file' rule='test_rule_name' matches=['$=test string'] tags=['tag1', 'tag2', 'tag3'] digest=(md5=d690ba32b59d28614aebefe9b03c74d4, sha1=4b1ced217aabe37138e96fb93bf40026639b9d3b, sha256=7a644118588ff0dcf2fadbe198ae1f1629c29374bac491ba41d5cf957edf0dfc)"  # noqa E501
        hit2 = "<filesystem/yara/match hostname=None domain=None path='/test/dir/to/test_file' rule='test_rule_name' matches=['$=test string'] tags=['tag1', 'tag2', 'tag3'] digest=(md5=bd7490dd2978ce983e2e1613ac8444c0, sha1=849a062cf09280f5c7dce4c7f87c69a1d9262e08, sha256=9bf7629a67c7ce8019910f1c1251fe44b61b3fff55a59a5e148af3c207dc102f)"  # noqa E501

        assert hit1 in out
        assert hit2 in out



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers import configutil
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.record import (
    COMMON_APPLICATION_FIELDS,
    TargetRecordDescriptor,
)
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target

UnixApplicationRecord = TargetRecordDescriptor(
    "unix/application",
    COMMON_APPLICATION_FIELDS,
)


class UnixApplicationsPlugin(Plugin):
    """Unix Applications plugin."""

    SYSTEM_PATHS = (
        "/usr/share/applications/",
        "/usr/local/share/applications/",
        "/var/lib/snapd/desktop/applications/",
        "/var/lib/flatpak/exports/share/applications/",
    )

    USER_PATHS = (".local/share/applications/",)

    SYSTEM_APPS = ("org.gnome.",)

    def __init__(self, target: Target):
        super().__init__(target)
        self.desktop_files = list(self._find_desktop_files())

    def _find_desktop_files(self) -> Iterator[TargetPath]:
        for dir in self.SYSTEM_PATHS:
            for file in self.target.fs.path(dir).glob("*.desktop"):
                yield file

        for user_details in self.target.user_details.all_with_home():
            for dir in self.USER_PATHS:
                for file in user_details.home_path.joinpath(dir).glob("*.desktop"):
                    yield file

    def check_compatible(self) -> None:
        if not self.desktop_files:
            raise UnsupportedPluginError("No application .desktop files found")

    @export(record=UnixApplicationRecord)
    def applications(self) -> Iterator[UnixApplicationRecord]:
        """Yield installed Unix GUI applications from GNOME and XFCE.

        References:
            - https://wiki.archlinux.org/title/Desktop_entries
            - https://specifications.freedesktop.org/desktop-entry-spec/latest/
            - https://unix.stackexchange.com/questions/582928/where-gnome-apps-are-installed

        Yields ``UnixApplicationRecord`` records with the following fields:

        .. code-block:: text

            ts_modified  (datetime): timestamp when the installation was modified
            ts_installed (datetime): timestamp when the application was installed on the system
            name         (string):   name of the application
            version      (string):   version of the application
            author       (string):   author of the application
            type         (string):   type of the application, either user or system
            path         (string):   path to the desktop file entry of the application
        """
        for file in self.desktop_files:
            config = configutil.parse(file, hint="ini").get("Desktop Entry") or {}
            stat = file.lstat()

            yield UnixApplicationRecord(
                ts_modified=stat.st_mtime,
                ts_installed=stat.st_btime if hasattr(stat, "st_btime") else None,
                name=config.get("Name"),
                version=config.get("Version"),
                path=config.get("Exec"),
                type="system" if config.get("Icon", "").startswith(self.SYSTEM_APPS) else "user",
                _target=self.target,
            )



from __future__ import annotations

import logging
import tarfile as tf
from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target import filesystem, target
from dissect.target.filesystems.tar import (
    TarFilesystemDirectoryEntry,
    TarFilesystemEntry,
)
from dissect.target.helpers import fsutil, loaderutil
from dissect.target.helpers.lazy import import_lazy
from dissect.target.loader import Loader, SubLoader

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

log = logging.getLogger(__name__)

TAR_EXT_COMP = (
    ".tar.gz",
    ".tar.xz",
    ".tar.bz",
    ".tar.bz2",
    ".tar.lzma",
    ".tar.lz",
    ".tgz",
    ".txz",
    ".tbz",
    ".tbz2",
    ".tlz",
    ".tlzma",
)
TAR_EXT = (".tar",)

TAR_MAGIC_COMP = (
    # gzip
    b"\x1f\x8b",
    # bzip2
    b"\x42\x5a\x68",
    # xz
    b"\xfd\x37\x7a\x58\x5a\x00",
    # lzma
    b"\x5d\x00\x00\x01\x00",
    b"\x5d\x00\x00\x10\x00",
    b"\x5d\x00\x00\x08\x00",
    b"\x5d\x00\x00\x10\x00",
    b"\x5d\x00\x00\x20\x00",
    b"\x5d\x00\x00\x40\x00",
    b"\x5d\x00\x00\x80\x00",
    b"\x5d\x00\x00\x00\x01",
    b"\x5d\x00\x00\x00\x02",
)
TAR_MAGIC = (tf.GNU_MAGIC, tf.POSIX_MAGIC)

WINDOWS_MEMBERS = (
    "windows/system32",
    "/windows/system32",
    "winnt",
    "/winnt",
)


class TarSubLoader(SubLoader[tf.TarFile]):
    """Tar implementation of a :class:`SubLoader`."""

    def __init__(self, path: Path, tar: tf.TarFile, **kwargs):
        super().__init__(path, tar, **kwargs)
        self.tar = tar

    @staticmethod
    def detect(path: Path, tarfile: tf.TarFile) -> bool:
        """Only to be called internally by :class:`TarLoader`."""
        raise NotImplementedError

    def map(self, target: target.Target) -> None:
        """Only to be called internally by :class:`TarLoader`."""
        raise NotImplementedError


class GenericTarSubLoader(TarSubLoader):
    """Generic tar sub loader."""

    @staticmethod
    def detect(path: Path, tarfile: tf.TarFile) -> bool:
        return True

    def map(self, target: target.Target) -> None:
        volumes = {}
        windows_found = False

        for member in self.tar.getmembers():
            if member.name == ".":
                continue

            if member.name.lower().startswith(WINDOWS_MEMBERS):
                windows_found = True
                if "/" in volumes:
                    # Root filesystem was already added
                    volumes["/"].case_sensitive = False

            if "/" not in volumes:
                vol = filesystem.VirtualFilesystem(case_sensitive=not windows_found)
                vol.tar = self.tar
                volumes["/"] = vol
                target.filesystems.add(vol)

            volume = volumes["/"]
            mname = member.name

            entry_cls = TarFilesystemDirectoryEntry if member.isdir() else TarFilesystemEntry
            entry = entry_cls(volume, fsutil.normpath(mname), member)
            volume.map_file_entry(entry.path, entry)

        for vol_name, vol in volumes.items():
            loaderutil.add_virtual_ntfs_filesystem(
                target,
                vol,
                usnjrnl_path=[
                    "$Extend/$Usnjrnl:$J",
                    "$Extend/$Usnjrnl:J",  # Old versions of acquire used $Usnjrnl:J
                ],
            )

            target.fs.mount(vol_name, vol)


class TarLoader(Loader):
    """Load tar files."""

    __subloaders__ = (
        import_lazy("dissect.target.loaders.containerimage").ContainerImageTarSubLoader,
        import_lazy("dissect.target.loaders.acquire").AcquireTarSubLoader,
        import_lazy("dissect.target.loaders.uac").UacTarSubloader,
        GenericTarSubLoader,  # should be last
    )

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)

        if is_compressed(path):
            log.warning(
                "Tar file %r is compressed, which will affect performance. "
                "Consider uncompressing the archive before passing the tar file to Dissect.",
                path,
            )

        self.fh = path.open("rb")
        self.tar = tf.open(mode="r:*", fileobj=self.fh)  # noqa: SIM115
        self.subloader = None

    @staticmethod
    def detect(path: Path) -> bool:
        return path.name.lower().endswith(TAR_EXT + TAR_EXT_COMP) or is_tar_magic(path, TAR_MAGIC + TAR_MAGIC_COMP)

    def map(self, target: target.Target) -> None:
        for candidate in self.__subloaders__:
            if candidate.detect(self.path, self.tar):
                self.subloader = candidate(self.path, self.tar, parsed_path=self.parsed_path)
                self.subloader.map(target)
                break


def is_tar_magic(path: Path, magics: Iterable[bytes]) -> bool:
    if not path.is_file():
        return False

    with path.open("rb") as fh:
        # The minimum file size of an uncompressed tar is 512 bytes, but a compressed tar file could be smaller.
        fh.seek(0)
        buf = fh.read(tf.BLOCKSIZE)
        headers = [buf[0:6]]
        if len(buf) >= 265:
            headers.append(buf[257 : 257 + 8])

        for header in headers:
            if header.startswith(magics):
                # We could be dealing with a compressed file that is not actually a tar.
                # To weed out a false positive we try to decompress and read ustar from
                # the first 512 bytes (or less) of the file.
                try:
                    tf.open(mode="r:*", fileobj=BytesIO(buf))  # noqa: SIM115
                except (tf.ReadError, tf.CompressionError, ValueError, EOFError):
                    continue
                return True
    return False


def is_compressed(path: Path) -> bool:
    return path.name.lower().endswith(TAR_EXT_COMP) or is_tar_magic(path, TAR_MAGIC_COMP)



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.generic import GenericPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target

log_mtime = absolute_path("_data/plugins/os/unix/log/empty.log").stat().st_mtime


def test_unix_generic_activity(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file("/var/log/some.log", absolute_path("_data/plugins/os/unix/log/empty.log"))
    target_unix.add_plugin(GenericPlugin)
    assert target_unix.activity == datetime.fromtimestamp(log_mtime, tz=timezone.utc)


def test_unix_generic_install_date(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file("/var/log/installer/syslog", absolute_path("_data/plugins/os/unix/log/empty.log"))
    target_unix.add_plugin(GenericPlugin)
    assert target_unix.install_date == datetime.fromtimestamp(log_mtime, tz=timezone.utc)



from __future__ import annotations

from dataclasses import dataclass, replace
from typing import TYPE_CHECKING, Any, Callable

from dissect.ntfs.c_ntfs import FILE_RECORD_SEGMENT_IN_USE
from flow.record.fieldtypes import windows_path

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export
from dissect.target.plugins.filesystem.ntfs.utils import (
    InformationType,
    get_drive_letter,
    get_owner_and_group,
    get_record_size,
    get_volume_identifier,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.ntfs import MftRecord
    from dissect.ntfs.attr import Attribute, FileName, StandardInformation
    from flow.record import Record
    from typing_extensions import Self

    from dissect.target.filesystems.ntfs import NtfsFilesystem
    from dissect.target.target import Target


FilesystemStdCompactRecord = TargetRecordDescriptor(
    "filesystem/ntfs/mft/std/compact",
    [
        ("datetime", "creation_time"),
        ("datetime", "last_modification_time"),
        ("datetime", "last_change_time"),
        ("datetime", "last_access_time"),
        ("uint32", "segment"),
        ("path", "path"),
        ("string", "owner"),
        ("filesize", "filesize"),
        ("boolean", "resident"),
        ("boolean", "inuse"),
        ("string", "volume_uuid"),
    ],
)


FilesystemStdRecord = TargetRecordDescriptor(
    "filesystem/ntfs/mft/std",
    [
        ("datetime", "ts"),
        ("string", "ts_type"),
        ("uint32", "segment"),
        ("path", "path"),
        ("string", "owner"),
        ("filesize", "filesize"),
        ("boolean", "resident"),
        ("boolean", "inuse"),
        ("string", "volume_uuid"),
    ],
)

FilesystemFilenameCompactRecord = TargetRecordDescriptor(
    "filesystem/ntfs/mft/filename/compact",
    [
        ("datetime", "creation_time"),
        ("datetime", "last_modification_time"),
        ("datetime", "last_change_time"),
        ("datetime", "last_access_time"),
        ("uint32", "filename_index"),
        ("uint32", "segment"),
        ("path", "path"),
        ("string", "owner"),
        ("filesize", "filesize"),
        ("boolean", "resident"),
        ("boolean", "inuse"),
        ("boolean", "ads"),
        ("string", "volume_uuid"),
    ],
)

FilesystemFilenameRecord = TargetRecordDescriptor(
    "filesystem/ntfs/mft/filename",
    [
        ("datetime", "ts"),
        ("string", "ts_type"),
        ("uint32", "filename_index"),
        ("uint32", "segment"),
        ("path", "path"),
        ("string", "owner"),
        ("filesize", "filesize"),
        ("boolean", "resident"),
        ("boolean", "inuse"),
        ("boolean", "ads"),
        ("string", "volume_uuid"),
    ],
)

FilesystemMACBRecord = TargetRecordDescriptor(
    "filesystem/ntfs/mft/macb",
    [
        ("datetime", "ts"),
        ("string", "macb"),
        ("uint32", "filename_index"),
        ("uint32", "segment"),
        ("path", "path"),
        ("string", "owner"),
        ("filesize", "filesize"),
        ("boolean", "resident"),
        ("boolean", "inuse"),
        ("boolean", "ads"),
        ("string", "volume_uuid"),
    ],
)

RECORD_TYPES = {
    InformationType.STANDARD_INFORMATION: FilesystemStdRecord,
    InformationType.FILE_INFORMATION: FilesystemFilenameRecord,
}


COMPACT_RECORD_TYPES = {
    InformationType.STANDARD_INFORMATION: FilesystemStdCompactRecord,
    InformationType.FILE_INFORMATION: FilesystemFilenameCompactRecord,
}

FORMAT_INFO = {
    InformationType.FILE_INFORMATION: ("F", ""),
    InformationType.STANDARD_INFORMATION: ("S", ""),
    InformationType.ALTERNATE_DATA_STREAM: ("F", " Is_ADS"),
}


class MftPlugin(Plugin):
    """NTFS MFT plugin."""

    __namespace__ = "mft"

    def __init__(self, target: Target):
        super().__init__(target)
        self.ntfs_filesystems = {index: fs for index, fs in enumerate(self.target.filesystems) if fs.__type__ == "ntfs"}

    def check_compatible(self) -> None:
        if not len(self.ntfs_filesystems):
            raise UnsupportedPluginError("No NTFS filesystems found")

    @export(
        record=[
            FilesystemStdRecord,
            FilesystemFilenameRecord,
            FilesystemStdCompactRecord,
            FilesystemFilenameCompactRecord,
        ]
    )
    @arg(
        "--compact",
        group="fmt",
        action="store_true",
        help="compacts the MFT entry timestamps into a single record",
    )
    @arg("--fs", type=int, help="optional filesystem index, zero indexed")
    @arg("--start", type=int, default=0, help="the first MFT segment number")
    @arg("--end", type=int, default=-1, help="the last MFT segment number")
    @arg(
        "--macb",
        group="fmt",
        action="store_true",
        help="compacts MFT timestamps into MACB bitfield (format: MACB[standard|ads]/MACB[filename])",
    )
    def records(
        self, compact: bool = False, fs: int | None = None, start: int = 0, end: int = -1, macb: bool = False
    ) -> Iterator[
        FilesystemStdRecord | FilesystemFilenameRecord | FilesystemStdCompactRecord | FilesystemFilenameCompactRecord
    ]:
        """Return the MFT records of all NTFS filesystems.

        The Master File Table (MFT) contains primarily metadata about every file and folder on a NFTS filesystem.

        If the filesystem is part of a virtual NTFS filesystem (a ``VirtualFilesystem`` with the MFT properties
        added to it through a "fake" ``NtfsFilesystem``), the paths returned in the MFT records are based on the
        mount point of the ``VirtualFilesystem``. This ensures that the proper original drive letter is used when
        available.
        When no drive letter can be determined, the path will show as e.g. ``\\$fs$\\fs0``.

        References:
            - https://docs.microsoft.com/en-us/windows/win32/fileio/master-file-table
        """

        record_formatter = default_formatter

        def noop_aggregator(records: Iterator[Record]) -> Iterator[Record]:
            yield from records

        aggregator = noop_aggregator

        if compact:
            record_formatter = compact_formatter
        elif macb:
            aggregator = macb_aggregator

        filesystems: list[NtfsFilesystem] = []
        if fs is not None:
            try:
                filesystems = [self.ntfs_filesystems[fs]]
            except KeyError:
                self.target.log.error("NTFS filesystem with index number %s does not exist", fs)  # noqa: TRY400
                return
        else:
            filesystems = self.ntfs_filesystems.values()

        for filesystem in filesystems:
            info = _Info.init(self.target, filesystem)

            try:
                for record in filesystem.ntfs.mft.segments(start, end):
                    try:
                        info.update(record, filesystem)

                        for path in record.full_paths():
                            path = f"{info.drive_letter}{path}"
                            yield from aggregator(
                                iter_records(
                                    record=record,
                                    segment=record.segment,
                                    path=path,
                                    owner=info.owner,
                                    size=info.size,
                                    resident=info.resident,
                                    inuse=info.in_use,
                                    drive_letter=info.drive_letter,
                                    volume_uuid=info.volume_uuid,
                                    record_formatter=record_formatter,
                                    target=self.target,
                                )
                            )
                    except Exception as e:  # noqa: PERF203
                        self.target.log.warning("An error occured parsing MFT segment %d: %s", record.segment, str(e))
                        self.target.log.debug("", exc_info=e)

            except Exception:
                self.target.log.exception("An error occured constructing FilesystemRecords")

    # Make calling the `mft` namespace backwards compatible with the old `mft` function
    __call__ = records

    @export(output="yield")
    @arg("--ignore-dos", action="store_true", help="ignore DOS file names")
    def timeline(self, ignore_dos: bool = False) -> Iterator[str]:
        """Return the MFT records of all NTFS filesystems in a human readable format (unsorted).

        The Master File Table (MFT) contains metadata about every file and folder on a NFTS filesystem.

        If the filesystem is part of a virtual NTFS filesystem (a ``VirtualFilesystem`` with the MFT properties
        added to it through a "fake" ``NtfsFilesystem``), the paths returned in the MFT records are based on the
        mount point of the ``VirtualFilesystem``. This ensures that the proper original drive letter is used when
        available.
        When no drive letter can be determined, the path will show as e.g. ``\\$fs$\\fs0``.

        References:
            - https://docs.microsoft.com/en-us/windows/win32/fileio/master-file-table
        """
        for fs in self.target.filesystems:
            if fs.__type__ != "ntfs":
                continue
            fs: NtfsFilesystem

            info = _Info.init(self.target, fs)

            for record in fs.ntfs.mft.segments():
                segment = record.segment

                try:
                    info.update(record, fs)

                    for path in record.full_paths(ignore_dos):
                        path = f"{info.drive_letter}{path}"

                        for attr in record.attributes.STANDARD_INFORMATION:
                            yield from format_timeline_info(
                                segment,
                                path,
                                info,
                                attr,
                                InformationType.STANDARD_INFORMATION,
                            )

                        for idx, attr in enumerate(record.attributes.FILE_NAME):
                            filepath = f"{info.drive_letter}{attr.full_path()}"

                            yield from format_timeline_info(
                                segment,
                                filepath,
                                info,
                                attr,
                                InformationType.FILE_INFORMATION,
                                idx=idx,
                            )

                        ads_extras = replace(info)
                        ads_info = record.attributes.FILE_NAME[0]

                        for attr in record.attributes.DATA:
                            if attr.name != "":  # ADS Data
                                ads_extras.resident = attr.resident
                                ads_extras.size = get_record_size(record, attr.name)

                                adspath = f"{path}:{attr.name}"

                                yield from format_timeline_info(
                                    segment,
                                    adspath,
                                    ads_extras,
                                    ads_info,
                                    InformationType.ALTERNATE_DATA_STREAM,
                                )
                except Exception as e:
                    self.target.log.warning("An error occured parsing MFT segment %d: %s", segment, str(e))
                    self.target.log.debug("", exc_info=e)

    @export(output="yield")
    def body(self) -> Iterator[str]:
        """Return the MFT records of all NTFS filesystems in bodyfile format.

        The file mode is not accurate. This value was only added to indicate
        if a record is a file or directory.

        The Master File Table (MFT) contains metadata about every file and folder on a NFTS filesystem.

        If the filesystem is part of a virtual NTFS filesystem (a ``VirtualFilesystem`` with the MFT properties
        added to it through a "fake" ``NtfsFilesystem``), the paths returned in the MFT records are based on the
        mount point of the ``VirtualFilesystem``. This ensures that the proper original drive letter is used when
        available.
        When no drive letter can be determined, the path will show as e.g. ``\\$fs$\\fs0``.

        References:
            - https://docs.microsoft.com/en-us/windows/win32/fileio/master-file-table
            - https://wiki.sleuthkit.org/index.php?title=Body_file
        """
        for fs in self.target.filesystems:
            if fs.__type__ != "ntfs":
                continue
            fs: NtfsFilesystem

            info = _Info.init(self.target, fs)

            for record in fs.ntfs.mft.segments():
                # Just to make it clear when something is a file or a dir.
                size = 0
                file_mode = "d/drwxrwxrwx"
                if not record.is_dir():
                    size = get_record_size(record)
                    file_mode = "r/rrwxrwxrwx"

                try:
                    for path in record.full_paths(False):
                        path = f"{info.drive_letter}{path}"

                        for attribute in record.attributes.STANDARD_INFORMATION:
                            yield format_body_info(
                                name=path,
                                inode=record.segment,
                                mode_as_string=file_mode,
                                size=size,
                                atime=int(attribute.last_access_time.timestamp()),
                                mtime=int(attribute.last_modification_time.timestamp()),
                                ctime=int(attribute.last_change_time.timestamp()),
                                crtime=int(attribute.creation_time.timestamp()),
                            )
                except Exception as e:
                    self.target.log.warning(
                        "An error occured parsing the $STANDARD_INFORMATION attribute of MFT segment %d: %s",
                        record.segment,
                        str(e),
                    )

                try:
                    for attribute in record.attributes.FILE_NAME:
                        path = f"{info.drive_letter}{attribute.full_path()} ($FILE_NAME)"  # fls like output
                        yield format_body_info(
                            name=path,
                            inode=record.segment,
                            mode_as_string=file_mode,
                            size=size,
                            atime=int(attribute.last_access_time.timestamp()),
                            mtime=int(attribute.last_modification_time.timestamp()),
                            ctime=int(attribute.last_change_time.timestamp()),
                            crtime=int(attribute.creation_time.timestamp()),
                        )
                except Exception as e:
                    self.target.log.warning(
                        "An error occured parsing the $FILE_NAME attribute of MFT segment %d: %s",
                        record.segment,
                        str(e),
                    )


def iter_records(
    record: MftRecord,
    segment: int,
    path: str,
    owner: str,
    size: int,
    resident: bool,
    inuse: bool,
    drive_letter: str,
    volume_uuid: str,
    record_formatter: Callable,
    target: Target,
) -> Iterator[Record]:
    for attr in record.attributes.STANDARD_INFORMATION:
        yield from record_formatter(
            attr=attr,
            record_type=InformationType.STANDARD_INFORMATION,
            segment=segment,
            path=windows_path(path),
            owner=owner,
            filesize=size,
            resident=resident,
            inuse=inuse,
            volume_uuid=volume_uuid,
            _target=target,
        )

    for idx, attr in enumerate(record.attributes.FILE_NAME):
        filepath = f"{drive_letter}{attr.full_path()}"

        yield from record_formatter(
            attr=attr,
            record_type=InformationType.FILE_INFORMATION,
            filename_index=idx,
            segment=segment,
            path=windows_path(filepath),
            owner=owner,
            filesize=size,
            resident=resident,
            ads=False,
            inuse=inuse,
            volume_uuid=volume_uuid,
            _target=target,
        )

    ads_attributes = (data_attr for data_attr in record.attributes.DATA if data_attr.name != "")
    ads_info = record.attributes.FILE_NAME[0]

    for data_attr in ads_attributes:
        resident = data_attr.resident
        size = get_record_size(record, data_attr.name)
        ads_path = f"{path}:{data_attr.name}"

        yield from record_formatter(
            attr=ads_info,
            record_type=InformationType.FILE_INFORMATION,
            filename_index=None,
            segment=segment,
            path=windows_path(ads_path),
            owner=owner,
            filesize=size,
            resident=resident,
            inuse=inuse,
            ads=True,
            volume_uuid=volume_uuid,
            _target=target,
        )


def compact_formatter(
    attr: Attribute, record_type: InformationType, **kwargs
) -> Iterator[FilesystemStdCompactRecord | FilesystemFilenameCompactRecord]:
    record_desc = COMPACT_RECORD_TYPES.get(record_type)
    yield record_desc(
        creation_time=attr.creation_time,
        last_modification_time=attr.last_modification_time,
        last_change_time=attr.last_change_time,
        last_access_time=attr.last_access_time,
        **kwargs,
    )


def default_formatter(
    attr: Attribute, record_type: InformationType, **kwargs
) -> Iterator[FilesystemStdRecord | FilesystemFilenameRecord]:
    record_desc = RECORD_TYPES.get(record_type)
    for type, timestamp in [
        ("B", attr.creation_time),
        ("C", attr.last_change_time),
        ("M", attr.last_modification_time),
        ("A", attr.last_access_time),
    ]:
        yield record_desc(ts=timestamp, ts_type=type, **kwargs)


def macb_aggregator(records: Iterator[Record]) -> Iterator[Record]:
    def macb_set(bitfield: str, index: int, letter: str) -> str:
        return bitfield[:index] + letter + bitfield[index + 1 :]

    macbs = []
    for record in records:
        found = False

        offset = 0
        if not getattr(record, "ads", False):
            offset = int(record._desc.name == "filesystem/ntfs/mft/filename") * 5

        field = "MACB".find(record.ts_type) + offset
        for macb in macbs:
            if macb.ts == record.ts and macb.path == record.path:
                macb.macb = macb_set(macb.macb, field, record.ts_type)
                found = True
                break

        if found:
            continue

        macb = FilesystemMACBRecord.init_from_record(record)
        macb.macb = "..../...."
        macb.macb = macb_set(macb.macb, field, record.ts_type)

        macbs.append(macb)

    yield from macbs


@dataclass
class _Info:
    in_use: bool | None = None
    resident: bool | None = None
    owner: str | None = None
    size: int | None = None
    serial: int | None = None
    volume_uuid: str | None = None
    drive_letter: str | None = None

    @classmethod
    def init(cls, target: Target, fs: NtfsFilesystem) -> Self:
        # If this filesystem is a "fake" NTFS filesystem, used to enhance a
        # VirtualFilesystem, The driveletter (more accurate mount point)
        # returned will be that of the VirtualFilesystem. This makes sure
        # the paths returned in the records are actually reachable.
        return cls(
            serial=fs.ntfs.serial,
            volume_uuid=get_volume_identifier(fs),
            drive_letter=get_drive_letter(target, fs),
        )

    def update(self, record: MftRecord, fs: NtfsFilesystem) -> None:
        in_use = bool(record.header.Flags & FILE_RECORD_SEGMENT_IN_USE)
        owner, _ = get_owner_and_group(record, fs)
        resident = None
        size = None

        if not record.is_dir():
            for data_attribute in record.attributes.DATA:
                if data_attribute.name == "":
                    resident = data_attribute.resident
                    break

            size = get_record_size(record)

        self.in_use = in_use
        self.resident = resident
        self.size = size
        self.owner = owner

    def format(self) -> str:
        return (
            f"InUse:{format_none_value(self.in_use)} "
            f"Resident:{format_none_value(self.resident)} "
            f"Owner:{format_none_value(self.owner)} "
            f"Size:{format_none_value(self.size)} "
            f"VolumeUUID:{format_none_value(self.volume_uuid)}"
        )


def format_timeline_info(
    segment: int,
    path: str,
    info: _Info,
    attr: FileName | StandardInformation,
    attr_type: InformationType,
    idx: str = "",
) -> Iterator[str]:
    start_letter, postfix = FORMAT_INFO.get(attr_type, ("", ""))
    timestamps = {
        "B": attr.creation_time,
        "C": attr.last_change_time,
        "M": attr.last_modification_time,
        "A": attr.last_access_time,
    }

    for ts_type, ts in timestamps.items():
        information_type = f"{start_letter}{idx}{ts_type}"
        base_info = f"{ts} {information_type} {segment} {path}"

        yield f"{base_info} - {info.format()}{postfix}"


def format_body_info(
    md5: str = "0",
    name: str = "0",
    inode: int = 0,
    mode_as_string: str = "0",
    uid: int = 0,
    gid: int = 0,
    size: int = 0,
    atime: int = 0,
    mtime: int = 0,
    ctime: int = 0,
    crtime: int = 0,
) -> str:
    return f"{md5}|{name}|{inode}|{mode_as_string}|{uid}|{gid}|{size}|{atime}|{mtime}|{ctime}|{crtime}"


def format_none_value(value: Any) -> str | Any:
    """Format the value if it is ``None``."""

    return value if value is not None else "No Data"



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.regf.trusteddocs import TrustedDocumentsPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_trusteddocs_plugin(target_win_users: Target, hive_hku: VirtualHive) -> None:
    """Test if we detect and parse TrustedDocs entries.

    References:
        - https://github.com/Psmths/windows-forensic-artifacts/blob/main/file-activity/trustrecords-registry-key.md
    """
    trusteddocs_key_name = "Software\\Microsoft\\Office\\16.0\\Word\\Security\\Trusted Documents"
    trusteddocs_key = VirtualKey(hive_hku, trusteddocs_key_name)

    subkey_name = "TrustRecords"
    subkey = VirtualKey(hive_hku, subkey_name)
    subkey.add_value(
        "%USERPROFILE%\\Downloads\\example1.docm",
        VirtualValue(
            hive_hku,
            "%USERPROFILE%\\Downloads\\example1.docm",
            bytes.fromhex("5AAC70BDC995DA010028A153C5FFFFFFA9DEB30101000000"),
        ),
    )
    subkey.add_value(
        "%USERPROFILE%\\Downloads\\example2.docm",
        VirtualValue(
            hive_hku,
            "%USERPROFILE%\\Downloads\\example2.docm",
            bytes.fromhex("5AAC70BDC995DA010028A153C5FFFFFFABDEB301FFFFFF7F"),
        ),
    )

    trusteddocs_key.add_subkey(subkey_name, subkey)
    hive_hku.map_key(trusteddocs_key_name, trusteddocs_key)

    target_win_users.add_plugin(TrustedDocumentsPlugin)

    results = list(target_win_users.trusteddocs())

    assert len(results) == 2

    assert results[0].raw == bytes.fromhex("5AAC70BDC995DA010028A153C5FFFFFFA9DEB30101000000")
    assert results[0].application == "Word"
    assert (
        results[0].document == "%USERPROFILE%\\Downloads\\example1.docm"
    )  # Resolve does not work here, should be 'C:\\Users\\John'
    assert results[0].ts_created == datetime(2024, 4, 23, 22, 1, 6, 9405, tzinfo=timezone.utc)
    assert results[0].ts_enabled == datetime(
        2024, 4, 23, 22, 1, tzinfo=timezone.utc
    )  # Unfortunately only minute-precision here.
    assert results[0].state == "EDITING_ENABLED"

    assert results[1].raw == bytes.fromhex("5AAC70BDC995DA010028A153C5FFFFFFABDEB301FFFFFF7F")
    assert results[1].application == "Word"
    assert (
        results[1].document == "%USERPROFILE%\\Downloads\\example2.docm"
    )  # Resolve does not work here, should be 'C:\\Users\\John'
    assert results[1].ts_created == datetime(2024, 4, 23, 22, 1, 6, 9405, tzinfo=timezone.utc)
    assert results[1].ts_enabled == datetime(
        2024, 4, 23, 22, 3, tzinfo=timezone.utc
    )  # Unfortunately only minute-precision here.
    assert results[1].state == "MACROS_ENABLED"



from __future__ import annotations

import argparse
import atexit
import logging
import math
import os
import re
import ssl
import sys
import time
import urllib.parse
from dataclasses import dataclass
from functools import lru_cache
from getpass import getpass
from pathlib import Path
from struct import pack, unpack_from
from threading import Thread
from typing import TYPE_CHECKING, Any, Callable, ClassVar, TypeVar

from dissect.util.stream import AlignedStream

from dissect.target.containers.raw import RawContainer
from dissect.target.exceptions import LoaderError
from dissect.target.loader import Loader
from dissect.target.plugin import arg

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

try:
    import paho.mqtt.client as mqtt

    HAS_PAHO = True
except ImportError:
    HAS_PAHO = False

log = logging.getLogger(__name__)

DISK_INDEX_OFFSET = 9

R = TypeVar("R")


def suppress(func: Callable[..., R]) -> Callable[..., R | None]:
    def suppressed(*args, **kwargs) -> Any:
        try:
            return func(*args, **kwargs)
        except Exception:
            return None

    return suppressed


@dataclass
class InfoMessage:
    disks: list[DiskMessage]


@dataclass
class DiskMessage:
    index: int = 0
    sector_size: int = 0
    total_size: int = 0


@dataclass
class SeekMessage:
    data: bytes = b""


class MqttTransferRatePerSecond:
    def __init__(self, window_size: int = 10):
        self.window_size = window_size
        self.timestamps = []
        self.bytes = []

    def record(self, timestamp: float, byte_count: int) -> MqttTransferRatePerSecond:
        while self.timestamps and (timestamp - self.timestamps[0] > self.window_size):
            self.timestamps.pop(0)
            self.bytes.pop(0)

        self.timestamps.append(timestamp)
        self.bytes.append(byte_count)
        return self

    def value(self, current_time: float) -> float:
        if not self.timestamps:
            return 0

        elapsed_time = current_time - self.timestamps[0]
        if elapsed_time == 0:
            return 0

        total_bytes = self.bytes[-1] - self.bytes[0]

        return total_bytes / elapsed_time


class MqttStream(AlignedStream):
    def __init__(self, stream: MqttConnection, disk_id: int, size: int | None = None):
        self.stream = stream
        self.disk_id = disk_id
        super().__init__(size)

    def _read(self, offset: int, length: int, optimization_strategy: int = 0) -> bytes:
        return self.stream.read(self.disk_id, offset, length, optimization_strategy)


class MqttDiagnosticLine:
    def __init__(self, connection: MqttConnection, total_peers: int):
        self.connection = connection
        self.total_peers = total_peers
        self._columns, self._rows = os.get_terminal_size(0)
        atexit.register(self._detach)
        self._attach()

    def _attach(self) -> None:
        # save cursor position
        sys.stderr.write("\0337")
        # set top and bottom margins of the scrolling region to default
        sys.stderr.write("\033[r")
        # restore cursor position
        sys.stderr.write("\0338")
        # move cursor down one line in the same column; if at the bottom, the screen scrolls up
        sys.stderr.write("\033D")
        # move cursor up one line in the same column; if at the top, screen scrolls down
        sys.stderr.write("\033M")
        # save cursor position again
        sys.stderr.write("\0337")
        # restrict scrolling to a region from the first line to one before the last line
        sys.stderr.write(f"\033[1;{self._rows - 1}r")
        # restore cursor position after setting scrolling region
        sys.stderr.write("\0338")

    def _detach(self) -> None:
        # save cursor position
        sys.stderr.write("\0337")
        # move cursor to the specified position (last line, first column)
        sys.stderr.write(f"\033[{self._rows};1H")
        # clear from cursor to end of the line
        sys.stderr.write("\033[K")
        # reset scrolling region to include the entire display
        sys.stderr.write("\033[r")
        # restore cursor position
        sys.stderr.write("\0338")
        # ensure the written content is displayed (flush output)
        sys.stderr.flush()

    def display(self) -> None:
        # prepare: set background color to blue and text color to white at the beginning of the line
        prefix = "\x1b[44m\x1b[37m\r"
        # reset all attributes (colors, styles) to their defaults afterwards
        suffix = "\x1b[0m"
        # separator to set background color to red and text style to bold
        separator = "\x1b[41m\x1b[1m"
        logo = "TARGETD"

        start = time.time()
        transfer_rate = MqttTransferRatePerSecond(window_size=7)

        while True:
            time.sleep(0.05)
            peers = "?"
            try:
                peers = len(self.connection.broker.peers(self.connection.host))
            except Exception:
                pass

            recv = self.connection.broker.bytes_received
            now = time.time()
            transfer = transfer_rate.record(now, recv).value(now) / 1000  # convert to KB/s
            failures = self.connection.retries
            seconds_elapsed = round(now - start) % 60
            minutes_elapsed = math.floor((now - start) / 60) % 60
            hours_elapsed = math.floor((now - start) / 60**2)
            timer = f"{hours_elapsed:02d}:{minutes_elapsed:02d}:{seconds_elapsed:02d}"
            display = f"{timer} {peers}/{self.total_peers} peers {transfer:>8.2f} KB p/s {failures:>4} failures"
            rest = self._columns - len(display)
            padding = (rest - len(logo)) * " "

            # save cursor position
            sys.stderr.write("\0337")
            # move cursor to specified position (last line, first column)
            sys.stderr.write(f"\033[{self._rows};1H")
            # disable line wrapping
            sys.stderr.write("\033[?7l")
            # reset all attributes
            sys.stderr.write("\033[0m")
            # write the display line with prefix, calculated display content, padding, separator, and logo
            sys.stderr.write(prefix + display + padding + separator + logo + suffix)
            # enable line wrapping again
            sys.stderr.write("\033[?7h")
            # restore cursor position
            sys.stderr.write("\0338")
            # flush output to ensure it is displayed
            sys.stderr.flush()

    def start(self) -> None:
        t = Thread(target=self.display)
        t.daemon = True
        t.start()


class MqttConnection:
    broker = None
    host = None
    prev = -1
    factor = 1
    prefetch_factor_inc = 10
    retries = 0

    def __init__(self, broker: Broker, host: str):
        self.broker = broker
        self.host = str(host)
        self.info = lru_cache(128)(self.info)
        self.read = lru_cache(128)(self.read)

    def topo(self, peers: int) -> list[str]:
        self.broker.topology(self.host)

        while len(self.broker.peers(self.host)) < peers:
            self.broker.topology(self.host)
            time.sleep(1)
        return self.broker.peers(self.host)

    def info(self) -> list[MqttStream]:
        disks = []
        self.broker.info(self.host)

        message = None
        while message is None:
            message = self.broker.disk(self.host)

        for idx, disk in enumerate(message.disks):
            disks.append(MqttStream(self, idx, disk.total_size))

        return disks

    def read(self, disk_id: int, offset: int, length: int, optimization_strategy: int) -> bytes:
        message = None

        message = self.broker.read(self.host, disk_id, offset, length)
        if message:
            return message.data

        if self.prev == offset - (length * self.factor):
            if self.factor < 500:
                self.factor += self.prefetch_factor_inc
        else:
            self.factor = 1

        self.prev = offset
        flength = length * self.factor
        self.broker.factor = self.factor
        self.broker.seek(self.host, disk_id, offset, flength, optimization_strategy)
        attempts = 0
        while True:
            if message := self.broker.read(self.host, disk_id, offset, length):
                # don't waste time with sleep if we have a response
                break

            attempts += 1
            time.sleep(0.1)
            if attempts > 300:
                # message might have not reached agent, resend...
                self.broker.seek(self.host, disk_id, offset, flength, optimization_strategy)
                attempts = 0
                self.retries += 1

        return message.data


class Broker:
    broker_host = None
    broker_port = None
    private_key_file = None
    certificate_file = None
    cacert_file = None
    mqtt_client = None
    connected = False
    case = None
    bytes_received = 0
    monitor = False
    username = None
    password = None

    diskinfo: ClassVar[dict] = {}
    index: ClassVar[dict] = {}
    topo: ClassVar[dict] = {}
    factor = 1

    def __init__(
        self, broker: str, port: str, key: str, crt: str, ca: str, case: str, username: str, password: str, **kwargs
    ):
        self.broker_host = broker
        self.broker_port = int(port)
        self.private_key_file = key
        self.certificate_file = crt
        self.cacert_file = ca
        self.case = case
        self.username = username
        self.password = password
        self.command = kwargs.get("command")

        if not HAS_PAHO:
            raise ImportError("Required dependency 'paho' is missing, install with 'pip install dissect.target[mqtt]'")

    def clear_cache(self) -> None:
        self.index = {}

    @suppress
    def read(self, host: str, disk_id: int, seek_address: int, read_length: int) -> SeekMessage:
        key = f"{host}-{disk_id}-{seek_address}-{read_length}"
        return self.index.get(key)

    @suppress
    def disk(self, host: str) -> DiskMessage:
        return self.diskinfo[host]

    def peers(self, host: str) -> list[str]:
        return self.topo[host]

    def _on_disk(self, hostname: str, payload: bytes) -> None:
        (num_of_disks,) = unpack_from("<B", payload, offset=0)
        disks = []
        for disk_index in range(num_of_disks):
            (
                sector_size,
                total_size,
            ) = unpack_from("<IQ", payload, offset=1 + (disk_index * DISK_INDEX_OFFSET))
            disks.append(DiskMessage(index=disk_index, sector_size=sector_size, total_size=total_size))

        self.diskinfo[hostname] = InfoMessage(disks=disks)

    def _on_read(self, hostname: str, tokens: list[str], payload: bytes) -> None:
        disk_id = tokens[3]
        seek_address = int(tokens[4], 16)
        read_length = int(tokens[5], 16)

        for i in range(self.factor):
            sublength = int(read_length / self.factor)
            start = i * sublength
            key = f"{hostname}-{disk_id}-{seek_address + start}-{sublength}"
            if key in self.index:
                continue

            self.index[key] = SeekMessage(data=payload[start : start + sublength])

    def _on_id(self, hostname: str, payload: bytes) -> None:
        key = hostname
        host = payload.decode("utf-8")
        if host not in self.topo[key]:
            self.topo[key].append(payload.decode("utf-8"))
            self.mqtt_client.subscribe(f"{self.case}/{host}/DISKS")
            self.mqtt_client.subscribe(f"{self.case}/{host}/READ/#")
            if self.command is not None:
                self.mqtt_client.subscribe(f"{self.case}/{host}/CALLID")
                self.mqtt_client.publish(f"{self.case}/{host}/COMM", self.command.encode("utf-8"))
            time.sleep(1)

    def _on_call_id(self, hostname: str, payload: bytes) -> None:
        try:
            decoded_payload = payload.decode("utf-8")
        except UnicodeDecodeError as e:
            log.exception("Failed to decode payload for hostname %s: %s", hostname, e)  # noqa: TRY401
            return

        print(decoded_payload)

    def _on_log(self, client: mqtt.Client, userdata: Any, log_level: int, message: str) -> None:
        log.debug(message)

    def _on_connect(self, client: mqtt.Client, userdata: Any, flags: dict, rc: int) -> None:
        self.connected = True

    def _on_message(self, client: mqtt.Client, userdata: Any, msg: mqtt.client.MQTTMessage) -> None:
        tokens = msg.topic.split("/")
        casename, hostname, response, *_ = tokens
        if casename != self.case:
            return

        if self.monitor:
            self.bytes_received += len(msg.payload)

        if response == "DISKS":
            self._on_disk(hostname, msg.payload)
        elif response == "READ":
            self._on_read(hostname, tokens, msg.payload)
        elif response == "ID":
            self._on_id(hostname, msg.payload)
        elif response == "CALLID":
            self._on_call_id(hostname, msg.payload)

    def seek(self, host: str, disk_id: int, offset: int, flength: int, optimization_strategy: int) -> None:
        length = int(flength / self.factor)
        key = f"{host}-{disk_id}-{offset}-{length}"
        if key in self.index:
            return

        self.mqtt_client.publish(
            f"{self.case}/{host}/SEEK/{disk_id}/{hex(offset)}/{hex(flength)}", pack("<I", optimization_strategy)
        )

    def info(self, host: str) -> None:
        self.mqtt_client.publish(f"{self.case}/{host}/INFO")

    def topology(self, host: str) -> None:
        if host not in self.topo:
            self.topo[host] = []
        self.mqtt_client.subscribe(f"{self.case}/{host}/ID")
        time.sleep(1)  # need some time to avoid race condition, i.e. MQTT might react too fast
        # send a simple clear command (invalid, just clears the prev. msg) just in case TOPO is stale
        self.mqtt_client.publish(f"{self.case}/{host}/CLR")
        self.mqtt_client.publish(f"{self.case}/{host}/TOPO")

    def connect(self) -> None:
        self.mqtt_client = mqtt.Client(
            client_id="", clean_session=True, userdata=None, protocol=mqtt.MQTTv311, transport="tcp"
        )
        self.mqtt_client.tls_set(
            ca_certs=self.cacert_file,
            certfile=self.certificate_file,
            keyfile=self.private_key_file,
            cert_reqs=ssl.CERT_REQUIRED,
            tls_version=ssl.PROTOCOL_TLS,
            ciphers=None,
        )
        self.mqtt_client.username_pw_set(self.username, self.password)
        self.mqtt_client.tls_insecure_set(True)  # merely having the correct cert is ok
        self.mqtt_client.on_connect = self._on_connect
        self.mqtt_client.on_message = self._on_message
        if log.getEffectiveLevel() == logging.DEBUG:
            self.mqtt_client.on_log = self._on_log
        self.mqtt_client.connect(self.broker_host, port=self.broker_port, keepalive=60)
        self.mqtt_client.loop_start()


def strictly_positive(value: str) -> int:
    """Validates that the provided value is a strictly positive integer.

    This function is intended to be used as a type for argparse arguments.

    Args:
        value (str): The value to validate.

    Returns:
        int: The validated integer value.

    Raises:
        argparse.ArgumentTypeError: If the value is not a strictly positive integer.
    """
    try:
        strictly_positive_value = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError(f"Invalid integer value specified: '{value}'")

    if strictly_positive_value < 1:
        raise argparse.ArgumentTypeError("Value must be larger than or equal to 1.")
    return strictly_positive_value


def port(value: str) -> int:
    """Convert a string value to an integer representing a valid port number.

    This function is intended to be used as a type for argparse arguments.

    Args:
        value (str): The string representation of the port number.
    Returns:
        int: The port number as an integer.
    Raises:
        argparse.ArgumentTypeError: If the port number is not an integer or out of the valid range (1-65535).
    """

    try:
        port = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError(f"Invalid port number specified: '{value}'")

    if port < 1 or port > 65535:
        raise argparse.ArgumentTypeError("Port number must be between 1 and 65535.")
    return port


def case(value: str) -> str:
    """Validates that the given value is a valid case name consisting of
    alphanumeric characters and underscores only.

    This function is intended to be used as a type for argparse arguments.

    Args:
        value (str): The case name to validate.

    Returns:
        str: The validated case name if it matches the required pattern.

    Raises:
        argparse.ArgumentTypeError: If the case name does not match the required pattern.
    """

    if re.match(r"^[a-zA-Z0-9_]+$", value):
        return value

    raise argparse.ArgumentTypeError(f"Invalid case name specified: '{value}'")


@arg(
    "--mqtt-peers",
    type=strictly_positive,
    dest="peers",
    default=1,
    help="minimum number of peers to await for first alias",
)
@arg(
    "--mqtt-case",
    type=case,
    dest="case",
    help="case name (broker will determine if you are allowed to access this data)",
)
@arg("--mqtt-port", type=port, dest="port", default=443, help="broker connection port")
@arg("--mqtt-broker", default="localhost", dest="broker", help="broker ip-address")
@arg("--mqtt-key", type=Path, dest="key", required=True, help="private key file")
@arg("--mqtt-crt", type=Path, dest="crt", required=True, help="client certificate file")
@arg("--mqtt-ca", type=Path, dest="ca", required=True, help="certificate authority file")
@arg("--mqtt-command", dest="command", help="direct command to client(s)")
@arg("--mqtt-diag", action="store_true", dest="diag", help="show MQTT diagnostic information")
@arg("--mqtt-username", dest="username", default="mqtt-loader", help="Username for connection")
@arg("--mqtt-password", action="store_true", dest="password", help="Ask for password before connecting")
class MqttLoader(Loader):
    """Load remote targets through a broker."""

    connection = None
    broker = None
    peers: ClassVar[list] = []

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs, resolve=False)
        cls = MqttLoader
        self.broker = cls.broker
        self.connection = MqttConnection(self.broker, path)

    @staticmethod
    def detect(path: Path) -> bool:
        return False

    @staticmethod
    def find_all(path: Path, parsed_path: urllib.parse.ParseResult | None = None) -> Iterator[str]:
        cls = MqttLoader
        num_peers = 1

        if parsed_path is None:
            raise LoaderError("No URI connection details have been passed.")

        if cls.broker is None:
            options = dict(urllib.parse.parse_qsl(parsed_path.query, keep_blank_values=True))
            if options.get("password"):
                options["password"] = getpass()
            cls.broker = Broker(**options)
            cls.broker.connect()
            num_peers = int(options.get("peers", 1))
            cls.connection = MqttConnection(cls.broker, path)
            if options.get("diag"):
                cls.broker.monitor = True
                MqttDiagnosticLine(cls.connection, num_peers).start()
        else:
            cls.connection = MqttConnection(cls.broker, path)

        cls.peers = cls.connection.topo(num_peers)
        yield from (f"mqtt://{peer}?{parsed_path.query}" for peer in cls.peers)

    def map(self, target: Target) -> None:
        for disk in self.connection.info():
            target.disks.add(RawContainer(disk))



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator


class ESXiChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields from ESXi VM inventory."""

    __type__ = "esxi"

    def check_compatible(self) -> None:
        if self.target.os != "esxi":
            raise UnsupportedPluginError("Not an ESXi operating system")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        for vm in self.target.vm_inventory():
            yield ChildTargetRecord(
                type=self.__type__,
                path=vm.path,
                _target=self.target,
            )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.archive import xva

from dissect.target.containers.raw import RawContainer
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class XvaLoader(Loader):
    """Load Citrix Hypervisor XVA format files.

    References:
        - https://docs.citrix.com/en-us/citrix-hypervisor/vms/import-export.html#xva-format
    """

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        self.xva = xva.XVA(path.open("rb"))

    @staticmethod
    def detect(path: Path) -> bool:
        return path.suffix.lower() == ".xva"

    def map(self, target: Target) -> None:
        for ref in self.xva.disks():
            disk = self.xva.open(ref)

            target.disks.add(RawContainer(disk))



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest
from asn1crypto.algos import EncryptionAlgorithmId
from flow.record.fieldtypes import datetime as dt

from dissect.target.helpers import keychain
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.plugins.apps.browser.firefox import (
    CKA_ID,
    FirefoxPlugin,
    decrypt,
    query_master_key,
    retrieve_master_key,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.fixture
def target_firefox_win(target_win_users: Target, fs_win: VirtualFilesystem) -> Target:
    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Mozilla\\Firefox\\Profiles\\g1rbw8y7.default-release\\",
        absolute_path("_data/plugins/apps/browser/firefox/"),
    )

    target_win_users.add_plugin(FirefoxPlugin)

    return target_win_users


@pytest.fixture
def target_firefox_unix(target_unix_users: Target, fs_unix: VirtualFilesystem) -> Target:
    fs_unix.map_dir(
        "/root/.mozilla/firefox/g1rbw8y7.default-release/", absolute_path("_data/plugins/apps/browser/firefox/")
    )
    fs_unix.map_dir(
        "/root/.mozilla/firefox/g1rbw8y7.default-release/",
        absolute_path("_data/plugins/apps/browser/firefox/passwords/default/"),
    )

    target_unix_users.add_plugin(FirefoxPlugin)

    return target_unix_users


@pytest.fixture
def target_firefox_oculus(target_android: Target, fs_android: VirtualFilesystem) -> Target:
    fs_android.map_dir(
        "/data/data/org.mozilla.vrbrowser",
        absolute_path("_data/plugins/apps/browser/firefox/android/org.mozilla.vrbrowser"),
    )

    target_android.add_plugin(FirefoxPlugin)

    return target_android


@pytest.mark.parametrize(
    ("target_platform", "expected_source"),
    [
        (
            "target_firefox_win",
            "C:\\Users\\John\\AppData\\local\\Mozilla\\Firefox\\Profiles\\g1rbw8y7.default-release\\places.sqlite",
        ),
        (
            "target_firefox_unix",
            "/root/.mozilla/firefox/g1rbw8y7.default-release/places.sqlite",
        ),
    ],
)
def test_firefox_history(target_platform: Target, expected_source: str, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.firefox.history())

    assert len(records) == 24
    assert {"firefox"} == {record.browser for record in records}

    assert records[0].ts == dt("2021-12-01T10:42:05.742000+00:00")
    assert records[0].browser == "firefox"
    assert records[0].id == 1
    assert records[0].url == "https://www.mozilla.org/privacy/firefox/"
    assert not records[0].title
    assert not records[0].description
    assert records[0].host == "www.mozilla.org"
    assert records[0].visit_type == 1
    assert records[0].visit_count == 1
    assert records[0].hidden
    assert not records[0].typed
    assert records[0].session == 0
    assert not records[0].from_visit
    assert not records[0].from_url
    assert records[0].source == expected_source


@pytest.mark.parametrize(
    "target_platform",
    ["target_firefox_win", "target_firefox_unix"],
)
def test_firefox_downloads(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.firefox.downloads())

    assert len(records) == 3
    assert {"firefox"} == {record.browser for record in records}

    assert records[0].id == 1
    assert records[0].ts_start == dt("2021-12-01T10:57:01.175000+00:00")
    assert records[0].ts_end == dt("2021-12-01T10:57:01.321000+00:00")
    assert (
        records[0].url
        == "https://dl.google.com/tag/s/appguid%3D%7B8A69D345-D564-463C-AFF1-A69D9E530F96%7D%26iid%3D%7B2098EF96-29DB"
        "-B268-0B90-01AD59CD5C17%7D%26lang%3Dnl%26browser%3D3%26usagestats%3D1%26appname%3DGoogle%2520Chrome%26needs"
        "admin%3Dprefers%26ap%3Dx64-stable-statsdef_1%26installdataindex%3Dempty/update2/installers/ChromeSetup.exe"
    )


@pytest.mark.parametrize(
    "target_platform",
    ["target_firefox_win", "target_firefox_unix"],
)
def test_firefox_cookies(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)

    records = list(target_platform.firefox.cookies())

    assert len(records) == 4
    assert {"firefox"} == {record.browser for record in records}

    assert records[0].ts_created == dt("2023-07-13 09:53:47.460676+00:00")
    assert sorted(c.name for c in records) == [
        "_lr_env_src_ats",
        "_lr_retry_request",
        "_uc_referrer",
        "_uc_referrer",
    ]


@pytest.mark.parametrize(
    "target_platform",
    ["target_firefox_win", "target_firefox_unix"],
)
def test_firefox_extensions(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)

    records = list(target_platform.firefox.extensions())

    assert {"firefox"} == {record.browser for record in records}
    assert len(records) == 2
    assert records[0].extension_id == "uBlock0@raymondhill.net"
    assert records[0].ts_install == dt("2024-04-23 07:07:21.516000+00:00")
    assert records[0].ts_update == dt("2024-04-23 07:07:21.516000+00:00")
    assert (
        records[0].ext_path == "C:\\Users\\Win11\\AppData\\Roaming\\Mozilla\\Firefox\\Profiles"
        "\\9nxit8q0.default-release\\extensions\\uBlock0@raymondhill.net.xpi"
    )
    assert records[0].permissions == [
        "alarms",
        "dns",
        "menus",
        "privacy",
        "storage",
        "tabs",
        "unlimitedStorage",
        "webNavigation",
        "webRequest",
        "webRequestBlocking",
    ]
    assert records[1].permissions == []


@pytest.mark.parametrize(
    "target_platform",
    ["target_firefox_win", "target_firefox_unix"],
)
def test_firefox_passwords(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)

    records = list(target_platform.firefox.passwords())
    assert len(records) == 2

    assert records[0].browser == "firefox"
    assert records[0].decrypted_username == "username"
    assert records[0].encrypted_password == bytes.fromhex(
        "303a0410f8000000000000000000000000000001301406082a864886f70d030704081d993c26413915090410d9f1d0595158040a88850f9bcfbd36cf"
    )
    assert records[0].decrypted_password == "password"

    assert records[1].browser == "firefox"
    assert records[1].decrypted_username == "username"
    assert records[1].encrypted_password == bytes.fromhex(
        "303a0410f8000000000000000000000000000001301406082a864886f70d030704086fc7f57e0d7c456d04109b40eb0ebf5275d3a3a26f46a910b975"
    )
    assert records[1].decrypted_password == "password"


def test_unix_firefox_passwords_with_primary_password(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_dir(
        "/root/.mozilla/firefox/g1rbw8y7.default-release/",
        absolute_path("_data/plugins/apps/browser/firefox/passwords/primary/"),
    )
    target_unix_users.add_plugin(FirefoxPlugin)

    keychain.register_key(
        keychain.KeyType.PASSPHRASE,
        "PrimaryPassword",
        identifier=None,
        provider="browser",
    )

    records = list(target_unix_users.firefox.passwords())

    assert len(records) == 1

    assert records[0].browser == "firefox"
    assert records[0].username == "root"
    assert records[0].user_home == "/root"
    assert records[0].decrypted_username == "username"
    assert records[0].encrypted_password == bytes.fromhex(
        "303a0410f8000000000000000000000000000001301406082a864886f70d03070408cd349ac9ccc45c950410724daaeb9d2f85491e4e1d08597ecf2d"
    )
    assert records[0].decrypted_password == "password"


def test_firefox_oculus_history(target_firefox_oculus: Target) -> None:
    records = list(target_firefox_oculus.firefox.history())

    assert len(records) == 191
    assert {"firefox"} == {record.browser for record in records}

    assert records[0].url == "https://webxr.today/"
    assert records[0].id == 1
    assert not records[0].description
    assert records[0].ts == dt("2021-11-04 13:29:30.780000+00:00")


def test_firefox_oculus_cookies(target_firefox_oculus: Target) -> None:
    records = list(target_firefox_oculus.firefox.cookies())

    assert len(records) == 33
    assert {"firefox"} == {record.browser for record in records}

    assert records[0].ts_created == dt("2023-07-06 07:50:55.352988+00:00")
    assert sorted(c.name for c in records) == [
        "AEC",
        "AWSALBTGCORS",
        "CONSENT",
        "Features",
        "Locale",
        "NID",
        "OptanonConsent",
        "SNID",
        "SOCS",
        "Session-Id",
        "Tag",
        "Tag",
        "Tag",
        "Tag",
        "VISITOR_INFO1_LIVE",
        "_ga",
        "_ga_18EC69JQ0P",
        "_ga_2VC139B3XV",
        "_gat_gtag_UA_36116321_2",
        "_gid",
        "_hjFirstSeen",
        "count",
        "is-proton-user",
        "lst",
        "lst",
        "mp_abe3945ad0ddaadc3d987393d8d7c2ce_mixpanel",
        "session_id",
        "spid.6b32",
        "spses.6b32",
        "visid_incap_1329355",
        "visid_incap_2188750",
        "visid_incap_2295456",
        "visid_incap_2809573",
    ]


def test_firefox_oculus_extensions(target_firefox_oculus: Target) -> None:
    records = list(target_firefox_oculus.firefox.extensions())

    assert {"firefox"} == {record.browser for record in records}
    assert len(records) == 5
    assert records[0].extension_id == "default-theme@mozilla.org"
    assert records[0].ts_install == dt("2021-11-04 13:29:29.988000+00:00")
    assert records[0].ts_update == dt("1970-01-01 00:00:00+00:00")
    assert records[0].ext_path is None
    assert records[0].permissions == []


@pytest.fixture
def path_key4(fs_unix: VirtualFilesystem) -> TargetPath:
    fs_unix.map_file(
        "/key4.db",
        absolute_path("_data/plugins/apps/browser/firefox/key4.db"),
    )
    return TargetPath(fs_unix, "/key4.db")


PRIMARY_PASSWORD = "PrimaryPassword"


@pytest.fixture
def path_key4_primary_password(fs_unix: VirtualFilesystem) -> TargetPath:
    fs_unix.map_file(
        "/key4.db",
        absolute_path("_data/plugins/apps/browser/firefox/passwords/primary/key4.db"),
    )
    return TargetPath(fs_unix, "/key4.db")


@pytest.fixture
def logins() -> dict[str, str]:
    return {
        "username": "MDoEEPgAAAAAAAAAAAAAAAAAAAEwFAYIKoZIhvcNAwcECIFVMX6MyYxpBBC2j8K+bCEaE9/FmqE1wo2A",
        "password": "MDoEEPgAAAAAAAAAAAAAAAAAAAEwFAYIKoZIhvcNAwcECB2ZPCZBORUJBBDZ8dBZUVgECoiFD5vPvTbP",
    }


@pytest.fixture
def logins_with_primary_password() -> dict[str, str]:
    return {
        "username": "MDoEEPgAAAAAAAAAAAAAAAAAAAEwFAYIKoZIhvcNAwcECPEZvPh6dhBkBBB3/O2puJy1NiBUo5gS8hZh",
        "password": "MDoEEPgAAAAAAAAAAAAAAAAAAAEwFAYIKoZIhvcNAwcECM00msnMxFyVBBByTarrnS+FSR5OHQhZfs8t",
    }


@pytest.fixture
def decrypted() -> dict[str, str]:
    return {
        "username": "username",
        "password": "password",
    }


def test_decrypt_is_succesful(path_key4: TargetPath, logins: dict[str, str], decrypted: dict[str, str]) -> None:
    dec_username, dec_password = decrypt(logins.get("username"), logins.get("password"), path_key4)
    assert dec_username == decrypted.get("username")
    assert dec_password == decrypted.get("password")


@patch("dissect.target.plugins.apps.browser.firefox.retrieve_master_key", side_effect=ValueError(""))
def test_decrypt_bad_master_key(path_key4: TargetPath, logins: dict[str, str]) -> None:
    with pytest.raises(ValueError, match="Failed to decrypt password using keyfile:"):
        decrypt(logins.get("username"), logins.get("password"), path_key4)


def test_decrypt_with_primary_password_is_succesful(
    path_key4_primary_password: TargetPath, logins_with_primary_password: dict[str, str], decrypted: dict[str, str]
) -> None:
    dec_username, dec_password = decrypt(
        logins_with_primary_password.get("username"),
        logins_with_primary_password.get("password"),
        path_key4_primary_password,
        PRIMARY_PASSWORD,
    )
    assert dec_username == decrypted.get("username")
    assert dec_password == decrypted.get("password")


def test_decrypt_with_bad_primary_password_is_unsuccesful(
    path_key4_primary_password: TargetPath,
    logins_with_primary_password: dict[str, str],
) -> None:
    with pytest.raises(
        ValueError, match="Failed to decrypt password using keyfile: /key4.db, password: BAD_PRIMARY_PASSWORD"
    ):
        decrypt(
            logins_with_primary_password.get("username"),
            logins_with_primary_password.get("password"),
            path_key4_primary_password,
            "BAD_PRIMARY_PASSWORD",
        )


def test_retrieve_master_key_is_succesful(path_key4: TargetPath) -> None:
    key, algorithm = retrieve_master_key(b"", path_key4)

    assert EncryptionAlgorithmId.map(algorithm) == "pbes2"
    assert key.hex() == "452c2f920285f794614af1c2d99ed331940d73298a526140"


@patch("dissect.target.plugins.apps.browser.firefox.query_master_key", return_value=(b"aaaa", "BAD_CKA_VALUE"))
def test_retrieve_master_key_bad_cka_value(mock_target: Target, path_key4: TargetPath) -> None:
    with pytest.raises(ValueError, match="Password master key CKA_ID 'BAD_CKA_VALUE' is not equal to expected value"):
        retrieve_master_key(b"", path_key4)


def test_query_master_key(path_key4: TargetPath) -> None:
    master_key, master_key_cka = query_master_key(path_key4)

    assert isinstance(master_key, bytes)
    assert len(master_key) == 148

    assert isinstance(master_key_cka, bytes)
    assert len(master_key_cka) == 16
    assert master_key_cka == CKA_ID
